"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests of the `magni.cs.indicators`.

Routine listings
----------------
CoherenceTest(unittest.TestCase)
    Tests of the calculate_coherence function.
MutualCoherenceTest(unittest.TestCase)
    Tests of the calculate_mutual_coherence function.
RelativeEnergyTest(unittest.TestCase)
    Tests of the calculate_relative_energy function.

"""

from __future__ import division
import unittest

import numpy as np

import magni


class CoherenceTest(unittest.TestCase):
    """
    Tests of the calculate_coherence function.

    """

    def setUp(self):
        np.random.seed(1512)

        self._Phi, self._Psi = _generate_test_problem(5, 10, range(0, 10, 2))
        self._M = self._Phi.dot(self._Psi)

        for i in range(10):
            self._M[:, i] = self._M[:, i] / np.linalg.norm(self._M[:, i])

        self._M = self._M.T.dot(self._M) - np.eye(10)
        self._func = magni.cs.indicators.calculate_coherence

    def test_default(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = {0: self._func(Phi, Psi, norm=0),
                1: self._func(Phi, Psi, norm=1),
                2: self._func(Phi, Psi, norm=2),
                np.inf: self._func(Phi, Psi, norm=np.inf)}
        test = self._func(Phi, Psi)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(test == true)

    def test_0(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.sum(np.abs(M) > 1e-9) / (M.size - M.shape[0])
        test = self._func(Phi, Psi, norm=0)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_1(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.sum(np.abs(M)) / (M.size - M.shape[0])
        test = self._func(Phi, Psi, norm=1)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_2(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = (np.sum(np.abs(M)**2) / (M.size - M.shape[0]))**(1 / 2)
        test = self._func(Phi, Psi, norm=2)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_inf(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.max(np.abs(M))
        test = self._func(Phi, Psi, norm=np.inf)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))


class MutualCoherenceTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(1514)

        self._Phi, self._Psi = _generate_test_problem(5, 10, range(0, 10, 2))
        self._M = self._Phi.dot(self._Psi)
        self._func = magni.cs.indicators.calculate_mutual_coherence

    def test_default(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = {0: self._func(Phi, Psi, norm=0),
                1: self._func(Phi, Psi, norm=1),
                2: self._func(Phi, Psi, norm=2),
                np.inf: self._func(Phi, Psi, norm=np.inf)}
        test = self._func(Phi, Psi)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(test == true)

    def test_0(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.sum(np.abs(M) > 1e-9) / M.size
        test = self._func(Phi, Psi, norm=0)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_1(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.sum(np.abs(M)) / M.size
        test = self._func(Phi, Psi, norm=1)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_2(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = (np.sum(np.abs(M)**2) / M.size)**(1 / 2)
        test = self._func(Phi, Psi, norm=2)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_inf(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M

        true = np.max(np.abs(M))
        test = self._func(Phi, Psi, norm=np.inf)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))


class RelativeEnergyTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(1510)

        self._Phi, self._Psi = _generate_test_problem(5, 10, range(0, 10, 2))
        self._M = self._Phi.dot(self._Psi)
        self._func = magni.cs.indicators.calculate_relative_energy

    def test_default(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M
        w = [np.linalg.norm(M[:, i]) for i in range(M.shape[1])]

        true = {'min': self._func(Phi, Psi, method='min'),
                'diff': self._func(Phi, Psi, method='diff'),
                'mean': self._func(Phi, Psi, method='mean'),
                'std': self._func(Phi, Psi, method='std')}
        test = self._func(Phi, Psi)

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(test == true)

    def test_min(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M
        w = [np.linalg.norm(M[:, i]) for i in range(M.shape[1])]

        true = np.min(w)
        test = self._func(Phi, Psi, method='min')

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_diff(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M
        w = [np.linalg.norm(M[:, i]) for i in range(M.shape[1])]

        true = np.max(w) - np.min(w)
        test = self._func(Phi, Psi, method='diff')

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_mean(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M
        w = [np.linalg.norm(M[:, i]) for i in range(M.shape[1])]

        true = np.mean(w)
        test = self._func(Phi, Psi, method='mean')

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))

    def test_std(self):
        Phi, Psi, M = self._Phi, self._Psi, self._M
        w = [np.linalg.norm(M[:, i]) for i in range(M.shape[1])]

        true = np.std(w)
        test = self._func(Phi, Psi, method='std')

        # print('Expected: {}'.format(true))
        # print('Received: {}'.format(test))

        self.assertTrue(np.allclose(test, true))


def _generate_test_problem(n, m, coords):
    Phi = np.zeros((n, m))

    for i, j in enumerate(coords):
        Phi[i, j] = 1

    Psi = np.random.randn(m, m)

    for i in range(m):
        Psi[:, i] = Psi[:, i] / np.linalg.norm(Psi[:, i])

    return Phi, Psi



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing a multi domain image class.

Routine listings
----------------
MultiDomainImage(object)
    Provide access to an image in the domains of a compressed sensing context.

"""

from __future__ import division

from magni.utils.matrices import Matrix as _Matrix
from magni.utils.matrices import MatrixCollection as _MatrixC
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


class MultiDomainImage(object):
    """
    Provide access to an image in the domains of a compressed sensing context.

    Given a measurement matrix and a dictionary, an image can be supplied in
    either the measurement domain, the image domain, or the coefficient domain.
    This class then provides access to the image in all three domains.

    Parameters
    ----------
    Phi : magni.utils.matrices.Matrix, magni.utils.matrices.MatrixCollection,
        or numpy.ndarray
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix, magni.utils.matrices.MatrixCollection,
        or numpy.ndarray
        The dictionary.

    Notes
    -----
    The image is only converted to other domains than the supplied when the
    the image is requested in another domain. The image is, however, stored in
    up to three versions internally in order to reduce computation overhead.
    This may introduce a memory overhead.

    Examples
    --------
    Define a measurement matrix which skips every other sample:

    >>> import numpy as np, magni
    >>> func = lambda vec: vec[::2]
    >>> func_T = lambda vec: np.float64([vec[0], 0, vec[1]]).reshape(3, 1)
    >>> Phi = magni.utils.matrices.Matrix(func, func_T, (), (2, 3))

    Define a dictionary which is simply a rotated identity matrix:

    >>> v = np.sqrt(0.5)
    >>> Psi = np.float64([[ v, -v,  0],
    ...                   [ v,  v,  0],
    ...                   [ 0,  0,  1]])

    Instantiate the current class to handle domains:

    >>> from magni.imaging.domains import MultiDomainImage
    >>> domains = MultiDomainImage(Phi, Psi)

    An image can the be supplied in any domain and likewise retrieved in any
    domain. For example, setting the measurements and getting the coefficients:

    >>> domains.measurements = np.ones(2).reshape(-1, 1)
    >>> np.set_printoptions(suppress=True)
    >>> print(domains.coefficients)
    [[ 0.70710678]
     [-0.70710678]
     [ 1.        ]]

    Or setting the coefficients and getting the image:

    >>> domains.coefficients = np.ones(3).reshape(-1, 1)
    >>> print(domains.image)
    [[ 0.        ]
     [ 1.41421356]
     [ 1.        ]]

    Or setting the image and getting the measurements:

    >>> domains.image = np.ones(3).reshape(-1, 1)
    >>> print(domains.measurements)
    [[ 1.]
     [ 1.]]

    """

    def __init__(self, Phi, Psi):
        @_decorate_validation
        def validate_input():
            _numeric('Phi', ('integer', 'floating', 'complex'), shape=(-1, -1))
            _numeric('Psi', ('integer', 'floating', 'complex'),
                     shape=(Phi.shape[1], -1))

        validate_input()

        self._Phi = Phi
        self._Psi = Psi
        self._has_data = False
        self._measurements = None
        self._image = None
        self._coefficients = None

    @property
    def coefficients(self):
        """
        Get the image in the coefficient domain.

        Returns
        -------
        coefficients : numpy.ndarray
            The dictionary coefficients of the image.

        """

        if self._has_data:
            if self._coefficients is None:
                if self._image is None:
                    self._image = self._Phi.T.dot(self._measurements)

                self._coefficients = self._Psi.T.dot(self._image)

            return self._coefficients

    @coefficients.setter
    def coefficients(self, value):
        """
        Set the image in the coefficient domain.

        Parameters
        ----------
        coefficients : numpy.ndarray
            The dictionary coefficients of the image.

        """

        @_decorate_validation
        def validate_input():
            _numeric('value', ('integer', 'floating', 'complex'),
                     shape=(self._Psi.shape[1], 1))

        validate_input()

        self._has_data = True
        self._measurements = None
        self._image = None
        self._coefficients = value

    @property
    def image(self):
        """
        Get the image in the image domain.

        Returns
        -------
        image : numpy.ndarray
            The image.

        """

        if self._has_data:
            if self._image is None:
                if self._measurements is not None:
                    self._image = self._Phi.T.dot(self._measurements)
                elif self._coefficients is not None:
                    self._image = self._Psi.dot(self._coefficients)

            return self._image

    @image.setter
    def image(self, value):
        """
        Set the image in the image domain.

        Parameters
        ----------
        image : numpy.ndarray
            The image.

        """

        @_decorate_validation
        def validate_input():
            _numeric('value', ('integer', 'floating', 'complex'),
                     shape=(self._Phi.shape[1], 1))

        validate_input()

        self._has_data = True
        self._measurements = None
        self._image = value
        self._coefficients = None

    @property
    def measurements(self):
        """
        Get the image in the measurement domain.

        Returns
        -------
        measurements : numpy.ndarray
            The measurements of the image.

        """

        if self._has_data:
            if self._measurements is None:
                if self._image is None:
                    self._image = self._Psi.dot(self._coefficients)

                self._measurements = self._Phi.dot(self._image)

            return self._measurements

    @measurements.setter
    def measurements(self, value):
        """
        Set the image in the measurement domain.

        Parameters
        ----------
        measurements : numpy.ndarray
            The measurements of the image.

        """

        @_decorate_validation
        def validate_input():
            _numeric('value', ('integer', 'floating', 'complex'),
                     shape=(self._Phi.shape[0], 1))

        validate_input()

        self._has_data = True
        self._measurements = value
        self._image = None
        self._coefficients = None



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing a CS reconstruction algorithm adapted configger subclass.

Routine listings
----------------
Configger(magni.utils.config.Configger)
    Provide functionality to access a set of configuration options.

Notes
-----
This module does not itself contain any configuration options and thus has no
access to any configuration options unlike the other config modules of `magni`.

"""

import numpy as np

from magni.utils.config import Configger as _Configger
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic


class Configger(_Configger):
    """
    Provide functionality to access a set of configuration options.

    The present class redefines the methods for retrieving configuration
    parameters in order to ensure the desired precision of the floating point
    parameter values.

    Parameters
    ----------
    params : dict
        The configuration options and their default values.
    valids : dict
        The validation schemes of the configuration options.

    Attributes
    ----------
    property

    See Also
    --------
    magni.utils.config.Configger : Superclass of the present class.

    """

    def __init__(self, params, valids):
        @_decorate_validation
        def validate_input():
            _generic('params', 'mapping', has_keys=('precision_float',))

        _Configger.__init__(self, params, valids)
        validate_input()

    def __getitem__(self, name):
        """
        Get the value of a configuration parameter.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        value : None
            The value of the parameter.

        Notes
        -----
        If the value is a floating point value then that value is typecast to
        the desired precision.

        """

        value = _Configger.__getitem__(self, name)

        if isinstance(value, (float, np.floating)):
            value = self['precision_float'](value)

        return value



#!/usr/bin/env python

"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Script for running all Magni tests using nosetests

This script identifies TestCases in modules in magni/magni/tests. It then runs
these TestCases in a temporary folder such that auxillary files created by the
tests are automatically removed after their have been run.

usage: run_tests.py [-h] [--keep-output] [--skip [SKIP [SKIP ...]]]

optional arguments:
  -h, --help                   show this help message and exit
  --keep-output                keep test output in the .magnitest directory
  --no-coverage                do not print a coverage report
  --skip [SKIP [SKIP ...]]     name(s) of test module(s) to skip

Additional arguments are passed on to the nose test runner.

"""

from __future__ import division, print_function
import argparse
import nose
import os
import shutil
import sys

try:
    _disp = os.environ['DISPLAY']
except KeyError:
    print('Warning: DISPLAY environment variable not set. ' +
          'Falling back to "Agg" backend for matplotlib.')
    import matplotlib as mpl
    mpl.use('Agg')


def run_tests(test_modules, args, nose_argv):
    """
    Run the tests using nosetests.

    Parameters
    ----------
    test_modules : list or tuple
        The modules (as strings) containing tests to run.
    args : list or tuple
        The no_coverage and keep_output argument flags.
    nose_argv : list or tuple
        The arguments to pass to nosetests.

    """

    # Create a temporary folder to run tests in
    tmp_test_dir = path + '.magnitests'
    print(('Creating temporary directory for test output: {}... ')
          .format(tmp_test_dir), end='')
    cur_dir = os.getcwd()
    try:
        os.mkdir(tmp_test_dir)
    except OSError:
        pass

    print('done')
    os.chdir(tmp_test_dir)

    # Run the tests according to specification
    print('Running tests defined in:')
    for module in test_modules:
        print('  - {}'.format(module))

    fixed_nose_argv = ['--no-byte-compile', '--nologcapture']

    if not args.no_coverage:
        fixed_nose_argv.extend(
            ['--with-coverage', '--cover-html', '--cover-package=magni',
             '--cover-branches'])

    nose_test_paths = [path + test_module for test_module in test_modules]

    success = nose.run(defaultTest=nose_test_paths,
                       argv=fixed_nose_argv + nose_argv)
    os.chdir(cur_dir)

    # Handle test output
    if args.keep_output:
        print('Test output may be found in: {}'.format(tmp_test_dir))
    else:
        print('Deleting temporary test output... ', end='')
        shutil.rmtree(tmp_test_dir)
        print('done')

    # Exit 0 on success as required by Travis-CI
    sys.exit(0 if success else 1)

if __name__ == '__main__':
    # Arguments parsing
    arg_parser = argparse.ArgumentParser(description='Magni test suite runner')
    arg_parser.add_argument('--keep-output', action='store_true',
                            help=('Keep test output in the' +
                                  '.magnitest directory'))
    arg_parser.add_argument('--no-coverage', action='store_true',
                            help=('Do not show coverage report'))
    arg_parser.add_argument('--skip', type=str, nargs='*', default=[],
                            help='Name(s) of test module(s) to skip')
    args, nose_argv = arg_parser.parse_known_args()

    # Identify modules containing tests
    excludes = ['run_tests.py', 'dep_check.py', '__init__.py'] + args.skip
    real_path = os.path.realpath(str(__file__))
    path_pos = str.rfind(real_path, os.path.sep) + 1
    path, name = real_path[:path_pos], real_path[path_pos:]
    test_modules = sorted([module for module in os.listdir(path)
                           if module[-3:] == '.py' and module not in excludes])

    run_tests(test_modules, args, nose_argv)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing 1D matrices for building 2D separable transforms.

Routine listings
----------------
get_DCT_transform_matrix(N)
    Return the normalised N-by-N discrete cosine transform (DCT) matrix.
get_DFT_transform_matrix(N)
    Return the normalised N-by-N discrete fourier transform (DFT) matrix.

"""

from __future__ import division

import numpy as np
import scipy.linalg

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def get_DCT_transform_matrix(N):
    """
    Return the normalised N-by-N discrete cosine transform (DCT) matrix.

    Applying the returned transform matrix to a vector x: D.dot(x) yields the
    DCT of x. Applying the returned transform matrix to a matrix A: D.dot(A)
    applies the DCT to the columns of A. Taking D.dot(A.dot(D.T)) applies the
    DCT to both columns and rows, i.e. a full 2D separable DCT transform. The
    inverse transform (the 1D IDCT) is D.T.

    Parameters
    ----------
    N : int
        The size of the DCT transform matrix to return.

    Returns
    -------
    D : ndarray
        The DCT transform matrix.

    Notes
    -----
    The returned DCT matrix normalised such that is consitutes a orthonormal
    transform as given by equations (2.119) and (2.120) in [1]_.

    References
    ----------
    .. [1] A.N. Akansu, R.A. Haddad, and P.R. Haddad, *Multiresolution Signal
       Decomposition: Transforms, Subbands, and Wavelets*, Academic Press,
       2000.

    Examples
    --------
    For example, get a 5-by-5 DCT matrix

    >>> import numpy as np
    >>> from magni.imaging.dictionaries import get_DCT_transform_matrix
    >>> D = get_DCT_transform_matrix(5)
    >>> np.round(np.abs(D), 4)
    array([[ 0.4472,  0.4472,  0.4472,  0.4472,  0.4472],
           [ 0.6015,  0.3717,  0.    ,  0.3717,  0.6015],
           [ 0.5117,  0.1954,  0.6325,  0.1954,  0.5117],
           [ 0.3717,  0.6015,  0.    ,  0.6015,  0.3717],
           [ 0.1954,  0.5117,  0.6325,  0.5117,  0.1954]])

    and apply the 2D DCT transform to a dummy image

    >>> np.random.seed(6021)
    >>> img = np.random.randn(5, 5)
    >>> img_dct = D.dot(img.dot(D.T))
    >>> np.round(img_dct, 4)
    array([[-0.5247, -0.0225,  0.9098,  0.369 , -0.477 ],
           [ 1.7309, -0.4142,  1.9455, -0.6726, -1.3676],
           [ 0.6987,  0.5355,  0.7213, -0.8498, -0.1023],
           [ 0.0078, -0.0545,  0.3649, -1.4694,  1.732 ],
           [-1.5864,  0.156 ,  0.8932, -0.8091,  0.5056]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('N', 'integer', range_='[1;inf)')

    validate_input()

    nn, rr = np.meshgrid(*map(np.arange, (N, N)))

    D = np.cos((2 * nn + 1) * rr * np.pi / (2 * N))
    D[0, :] /= np.sqrt(N)
    D[1:, :] /= np.sqrt(N/2)

    return D


def get_DFT_transform_matrix(N):
    """
    Return the normalised N-by-N discrete fourier transform (DFT) matrix.

    Applying the returned transform matrix to a vector x: D.dot(x) yields the
    DFT of x. Applying the returned transform matrix to a matrix A: D.dot(A)
    applies the DFT to the columns of A. Taking D.dot(A.dot(D.T)) applies the
    DFT to both columns and rows, i.e. a full 2D separable DFT transform. The
    inverse transform (the 1D IDFT) is D.T.

    Parameters
    ----------
    N : int
        The size of the DFT transform matrix to return.

    Returns
    -------
    D : ndarray
        The DFT transform matrix.

    See Also
    --------
    scipy.linalg.dft : The function used to generate the DFT transform matrix.

    Notes
    -----
    The returned DFT matrix normalised such that is consitutes a orthonormal
    transform as given by equations (2.105) and (2.109) in [2]_.

    References
    ----------
    .. [2] A.N. Akansu, R.A. Haddad, and P.R. Haddad, *Multiresolution Signal
       Decomposition: Transforms, Subbands, and Wavelets*, Academic Press,
       2000.

    Examples
    --------
    For example, get a 5-by-5 DFT matrix

    >>> import numpy as np, scipy.fftpack
    >>> from magni.imaging.dictionaries import get_DFT_transform_matrix
    >>> D = get_DFT_transform_matrix(5)
    >>> np.round(D, 2)
    array([[ 0.45+0.j  ,  0.45+0.j  ,  0.45+0.j  ,  0.45+0.j  ,  0.45+0.j  ],
           [ 0.45+0.j  ,  0.14-0.43j, -0.36-0.26j, -0.36+0.26j,  0.14+0.43j],
           [ 0.45+0.j  , -0.36-0.26j,  0.14+0.43j,  0.14-0.43j, -0.36+0.26j],
           [ 0.45+0.j  , -0.36+0.26j,  0.14-0.43j,  0.14+0.43j, -0.36-0.26j],
           [ 0.45+0.j  ,  0.14+0.43j, -0.36+0.26j, -0.36-0.26j,  0.14-0.43j]])

    and apply the 2D DFT transform to a dummy image

    >>> np.random.seed(6021)
    >>> img = np.random.randn(5, 5)
    >>> img_dft = D.dot(img.dot(D.T))
    >>> np.round(img_dft, 2)
    array([[-0.52+0.j  ,  0.44+0.48j,  0.11-0.39j,  0.11+0.39j,  0.44-0.48j],
           [ 1.04-0.59j,  1.32+0.13j, -1.20-0.39j,  0.35+0.66j,  0.19-0.36j],
           [ 0.18-1.24j,  0.75+0.44j, -0.75+0.72j, -0.52-0.8j ,  0.77+0.13j],
           [ 0.18+1.24j,  0.77-0.13j, -0.52+0.8j , -0.75-0.72j,  0.75-0.44j],
           [ 1.04+0.59j,  0.19+0.36j,  0.35-0.66j, -1.20+0.39j,  1.32-0.13j]])

    which may be shifted to have the zero-frequency component at the center of
    the spectrum

    >>> np.round(scipy.fftpack.fftshift(img_dft), 2)
    array([[-0.75-0.72j,  0.75-0.44j,  0.18+1.24j,  0.77-0.13j, -0.52+0.8j ],
           [-1.20+0.39j,  1.32-0.13j,  1.04+0.59j,  0.19+0.36j,  0.35-0.66j],
           [ 0.11+0.39j,  0.44-0.48j, -0.52+0.j  ,  0.44+0.48j,  0.11-0.39j],
           [ 0.35+0.66j,  0.19-0.36j,  1.04-0.59j,  1.32+0.13j, -1.20-0.39j],
           [-0.52-0.8j ,  0.77+0.13j,  0.18-1.24j,  0.75+0.44j, -0.75+0.72j]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('N', 'integer', range_='[1;inf)')

    validate_input()

    D = scipy.linalg.dft(N, scale='sqrtn')

    return D



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.utils.matrices`.

"""

from __future__ import division
import unittest

import numpy as np

import magni


class TestMatrix(unittest.TestCase):
    """
    Test of Matrix.

    Implemented tests:

    * test_asarray : transforming to an explicit ndarray
    * test_invalid_matrix : invalid complex conj or T matrix
    * test_matrix_state : return internal matrix state

    """

    def test_asarray(self):
        func = lambda vec: -vec
        matrix = magni.utils.matrices.Matrix(func, func, (), (3, 3))
        matrix_as_ndarray = np.asarray(matrix)
        self.assertTrue(np.allclose(matrix.A, matrix_as_ndarray))

    def test_invalid_matrix(self):
        vector = np.ones((3, 1))
        func = lambda vec: -vec
        complex_matrix = magni.utils.matrices.Matrix(
            func, func, (), (3, 3), is_complex=True)

        self.assertTrue(np.allclose(complex_matrix.dot(vector), -vector))
        self.assertTrue(
            np.allclose(complex_matrix.conj().T.dot(vector), -vector))
        self.assertTrue(np.allclose(complex_matrix.A, -np.eye(3)))

        with self.assertRaises(ValueError):
            complex_matrix.T.dot(vector)
        with self.assertRaises(ValueError):
            complex_matrix.conj().dot(vector)
        with self.assertRaises(ValueError):
            complex_matrix.T.A
        with self.assertRaises(ValueError):
            complex_matrix.conj().A

    def test_matrix_state(self):
        func = lambda vec: -vec
        conj_trans = lambda vec: vec
        matrix = magni.utils.matrices.Matrix(func, conj_trans, (), (3, 3))
        state = matrix.matrix_state
        vec = np.arange(3)

        self.assertTrue(np.allclose(state['func'](vec), func(vec)))
        self.assertTrue(np.allclose(state['conj_trans'](vec), conj_trans(vec)))
        self.assertEqual(len(state['args']), 0)
        self.assertFalse(state['is_complex'])
        self.assertTrue(state['is_valid'])


class TestMatrixCollection(unittest.TestCase):
    """
    Test of MatrixCollection.

    Implemented tests:

    * test_asarray : transforming to an explicit ndarray
    * test_incompatible_matrices : matrices with incompatible shapes
    * test_matrix_state : return internal matrix state

    """

    def test_asarray(self):
        func1 = lambda vec: -vec
        matrix1 = magni.utils.matrices.Matrix(func1, func1, (), (3, 3))
        func2 = lambda vec: vec[::-1]
        matrix2 = magni.utils.matrices.Matrix(func2, func2, (), (3, 3))
        matrix_collection = magni.utils.matrices.MatrixCollection(
            (matrix1, matrix2))
        matrix_collection_as_ndarray = np.asarray(matrix_collection)
        self.assertTrue(
            np.allclose(matrix_collection.A, matrix_collection_as_ndarray))

    def test_incompatible_matrices(self):
        func = lambda vec: -vec
        matrix1 = magni.utils.matrices.Matrix(func, func, (), (3, 3))
        matrix2 = magni.utils.matrices.Matrix(func, func, (), (4, 4))

        with self.assertRaises(ValueError):
            matrix_collection = magni.utils.matrices.MatrixCollection(
                (matrix1, matrix2))

    def test_matrix_state(self):
        func1 = lambda vec: -vec
        func2 = lambda vec: vec
        matrix1 = magni.utils.matrices.Matrix(func1, func1, (), (3, 3))
        matrix2 = magni.utils.matrices.Matrix(func2, func2, (), (3, 3))
        matrix_collection = magni.utils.matrices.MatrixCollection(
            [matrix1, matrix2])
        state = matrix_collection.matrix_state
        vec = np.arange(3).reshape(3, 1)

        self.assertTrue(np.allclose(state['matrices'][0].dot(vec), func1(vec)))
        self.assertTrue(np.allclose(state['matrices'][1].dot(vec), func2(vec)))

        mc_matrices = matrix_collection._matrices
        self.assertTrue(state['matrices'] is not mc_matrices)
        self.assertTrue(state['matrices'][0] is not mc_matrices[0])
        self.assertTrue(state['matrices'][1] is not mc_matrices[1])


class TestSeparable2DTransform(unittest.TestCase):
    """
    Test of Separable2DTransform.

    Implemented tests:

    * test_asarray : transforming to an explicit ndarray
    * test_DCT : comparison with FFT based DCT
    * test_DFT : comparison with FFT based DFT
    * test_matrix_state : return internal matrix state

    """

    def setUp(self):
        np.random.seed(6021)
        self.N = 2**8
        self.vec = magni.imaging.mat2vec(np.random.randn(self.N, self.N))
        self.mtx = np.random.randn(10, 10)

    def test_asarray(self):
        sep_matrix = magni.utils.matrices.Separable2DTransform(
            self.mtx, self.mtx)
        sep_matrix_as_ndarray = np.asarray(sep_matrix)
        self.assertTrue(np.allclose(sep_matrix.A, sep_matrix_as_ndarray))

    def test_DCT(self):
        dct_mtx = magni.imaging.dictionaries.get_DCT_transform_matrix(self.N)
        dct_matrix_sep = magni.utils.matrices.Separable2DTransform(
            dct_mtx.T, dct_mtx.T)
        dct_matrix_fft = magni.imaging.dictionaries.get_DCT((self.N, self.N))

        sep_dct = dct_matrix_sep.T.dot(self.vec)
        fft_dct = dct_matrix_fft.T.dot(self.vec)

        sep_idct = dct_matrix_sep.dot(self.vec)
        fft_idct = dct_matrix_fft.dot(self.vec)

        self.assertTrue(np.allclose(sep_dct, fft_dct))
        self.assertTrue(np.allclose(sep_idct, fft_idct))

    def test_DFT(self):
        dft_mtx = magni.imaging.dictionaries.get_DFT_transform_matrix(self.N)
        dft_matrix_sep = magni.utils.matrices.Separable2DTransform(
            dft_mtx.conj().T, dft_mtx.conj().T)
        dft_matrix_fft = magni.imaging.dictionaries.get_DFT((self.N, self.N))

        sep_dft = dft_matrix_sep.conj().T.dot(self.vec)
        fft_dft = dft_matrix_fft.conj().T.dot(self.vec)

        sep_idft = dft_matrix_sep.dot(self.vec)
        fft_idft = dft_matrix_fft.dot(self.vec)

        self.assertTrue(np.allclose(sep_dft, fft_dft))
        self.assertTrue(np.allclose(sep_idft, fft_idft))

    def test_matrix_state(self):
        sep_matrix = magni.utils.matrices.Separable2DTransform(
            self.mtx, self.mtx)
        state = sep_matrix.matrix_state
        vec = np.arange(100).reshape(100, 1)

        self.assertTrue(np.allclose(
            np.kron(state['mtx_l'], state['mtx_r']).dot(vec),
            sep_matrix.dot(vec)))

        self.assertTrue(state['mtx_l'] is not self.mtx)
        self.assertTrue(state['mtx_r'] is not self.mtx)


class TestNorm(unittest.TestCase):
    """
    Test of norm computation.

    Implemented tests:

    * test_Matrix : norm of Matrix
    * test_MatrixCollection : norm of MatrixCollection
    * test_Separable2DTransform_frobenius : Frobenius norm of 2D sep transform

    """

    def setUp(self):
        self.N = 10
        np.random.seed(6021)

    def test_Matrix(self):
        func = lambda vec: -vec
        matrix = magni.utils.matrices.Matrix(func, func, (), (3, 3))
        norm = magni.utils.matrices.norm(matrix)
        self.assertEqual(norm, np.sqrt(3))

    def test_MatrixCollection(self):
        func1 = lambda vec: -vec
        matrix1 = magni.utils.matrices.Matrix(func1, func1, (), (3, 3))
        func2 = lambda vec: vec[::-1]
        matrix2 = magni.utils.matrices.Matrix(func2, func2, (), (3, 3))
        matrix_collection = magni.utils.matrices.MatrixCollection(
            (matrix1, matrix2))
        norm = magni.utils.matrices.norm(matrix_collection)
        self.assertEqual(norm, np.sqrt(3))

    def test_Separable2DTransform_frobenius(self):
        mtx = np.random.randn(self.N, self.N)
        sep_matrix = magni.utils.matrices.Separable2DTransform(mtx, mtx)
        ref_norm = np.linalg.norm(sep_matrix.A)

        # Offload to numpy
        np_norm = np.linalg.norm(sep_matrix)
        self.assertEqual(np_norm, ref_norm)

        # Default none == fro norm
        none_norm = magni.utils.matrices.norm(sep_matrix)
        self.assertAlmostEqual(none_norm, ref_norm)

        # Direct fro norm
        fro_norm = magni.utils.matrices.norm(sep_matrix, ord='fro')
        self.assertAlmostEqual(fro_norm, ref_norm)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality related to linear transformations.

Routine listings
----------------
dct2(x, mn_tuple, overcomplete_mn_tuple=None)
    2D discrete cosine transform.
idct2(x, mn_tuple, overcomplete_mn_tuple=None)
    2D inverse discrete cosine transform.
dft2(x, mn_tuple, overcomplete_mn_tuple=None)
    2D discrete Fourier transform.
idft2(x, mn_tuple, overcomplete_mn_tuple=None)
    2D inverse discrete Fourier transform.

"""

from __future__ import division

import numpy as np
import scipy.fftpack

from magni.imaging._util import mat2vec as _mat2vec
from magni.imaging._util import vec2mat as _vec2mat
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def dct2(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Apply the 2D Discrete Cosine Transform (DCT).

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the matrix represented by `x`; `n`
        number of columns in the matrix.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the matrix represented by the
        function's output; `no` number of columns in the matrix.

    Returns
    -------
    alpha : ndarray
        When `overcomplete_mn_tuple` is omitted: an m*n x 1 vector of
        coefficients scaled such that x = idct2(dct2(x)). When
        `overcomplete_mn_tuple` is supplied: an mo*no x 1 vector of
        coefficients scaled such that x = idct2(dct2(x)).

    See Also
    --------
    scipy.fftpack.dct : 1D DCT

    Notes
    -----
    The overcomplete transform feature invoked by supplying
    `overcomplete_mn_tuple` is only available for SciPy >= 0.16.1.

    The 2D DCT uses the seperability property of the 1D DCT.
    http://stackoverflow.com/questions/14325795/scipys-fftpack-dct-and-idct
    Including the reshape operation, the full transform is:
    dct(dct(x.reshape(n, m).T).T).T.T.reshape(m * n, 1)

    """

    @_decorate_validation
    def validate_input():
        _validate_transform_fwd(x, mn_tuple, overcomplete_mn_tuple)

    validate_input()

    m, n = mn_tuple
    result = _vec2mat(x, (m, n))

    if overcomplete_mn_tuple is None:
        result = scipy.fftpack.dct(result, norm='ortho').T
        result = scipy.fftpack.dct(result, norm='ortho').T
    else:
        mo, no = overcomplete_mn_tuple
        result = scipy.fftpack.dct(result, norm='ortho', n=no).T
        result = scipy.fftpack.dct(result, norm='ortho', n=mo).T

    return _mat2vec(result)


def idct2(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Apply the 2D Inverse Discrete Cosine Transform (iDCT).

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The vector representing the associated column-stacked matrix. When
        `overcomplete_mn_tuple` is omitted: the shape must be m*n x 1. When
        `overcomplete_mn_tuple` is supplied: the shape must be mo*no x 1.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the associated matrix, `n` number of
        columns in the associated matrix. When `overcomplete_mn_tuple` is
        supplied, this is the shape of the function's output.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the associated matrix, `no` number
        of columns in the associated matrix. When supplied, this is the shape
        of the function's input.

    Returns
    -------
    ndarray
        An m*n x 1 vector of coefficients scaled such that x = dct2(idct2(x)).

    See Also
    --------
    scipy.fftpack.idct : 1D inverse DCT

    Notes
    -----
    The overcomplete transform feature invoked by supplying
    `overcomplete_mn_tuple` is only available for SciPy >= 0.16.1.

    """

    @_decorate_validation
    def validate_input():
        _validate_transform_bwd(x, mn_tuple, overcomplete_mn_tuple)

    validate_input()

    m, n = mn_tuple

    if overcomplete_mn_tuple is None:
        result = _vec2mat(x, (m, n))
        result = scipy.fftpack.idct(result, norm='ortho').T
        result = scipy.fftpack.idct(result, norm='ortho').T
    else:
        mo, no = overcomplete_mn_tuple
        result = _vec2mat(x, (mo, no))
        result = scipy.fftpack.idct(result, norm='ortho').T
        result = result[:n, :]
        result = scipy.fftpack.idct(result, norm='ortho').T
        result = result[:m, :]

    return _mat2vec(result)


def dft2(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Apply the 2D Discrete Fourier Transform (DFT).

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the matrix represented by `x`; `n`
        number of columns in the matrix.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the matrix represented by the
        function's output; `no` number of columns in the matrix.

    Returns
    -------
    ndarray
        When `overcomplete_mn_tuple` is omitted: an m*n x 1 vector of
        coefficients scaled such that x = idft2(dft2(x)). When
        `overcomplete_mn_tuple` is supplied: an mo*no x 1 vector of
        coefficients scaled such that x = idft2(dft2(x)).

    See Also
    --------
    numpy.fft.fft2 : The underlying 2D FFT used to compute the 2D DFT.

    Notes
    -----
    This is a normalised DFT, i.e. a normalisation constant of
    :math:`\sqrt{m * n}` is used.

    """

    @_decorate_validation
    def validate_input():
        _validate_transform_fwd(x, mn_tuple, overcomplete_mn_tuple)

    validate_input()

    m, n = mn_tuple
    result = _vec2mat(x, (m, n))

    if overcomplete_mn_tuple is None:
        result = 1 / np.sqrt(n * m) * np.fft.fft2(result)
    else:
        mo, no = overcomplete_mn_tuple
        result = 1 / np.sqrt(no * mo) * np.fft.fft2(result, s=(mo, no))

    return _mat2vec(result)


def idft2(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Apply the 2D Inverse Discrete Fourier Transform (iDFT).

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the associated matrix, `n` number of
        columns in the associated matrix. When `overcomplete_mn_tuple`
        is supplied, this is the shape of the function's output.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the associated matrix, `no` number
        of columns in the associated matrix. When supplied, this is the shape
        of the function's input.

    Returns
    -------
    ndarray
        An m*n x 1 vector of coefficients scaled such that x = dft2(idft2(x)).

    See Also
    --------
    numpy.fft.ifft2 : The underlying 2D iFFT used to compute the 2D iDFT.

    Notes
    -----
    This is a normalised iDFT, i.e. a normalisation constant of
    :math:`\sqrt{m * n}` is used.

    """

    @_decorate_validation
    def validate_input():
        _validate_transform_bwd(x, mn_tuple, overcomplete_mn_tuple)

    validate_input()

    m, n = mn_tuple

    if overcomplete_mn_tuple is None:
        result = _vec2mat(x, (m, n))
        result = np.sqrt(n * m) * np.fft.ifft2(result)
    else:
        mo, no = overcomplete_mn_tuple
        result = _vec2mat(x, (mo, no))
        result = np.sqrt(no * mo) * np.fft.ifft2(result)
        result = result[:m, :n]

    if np.allclose(result.imag, 0):
        result = result.real

    return _mat2vec(result)


def _validate_transform_fwd(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Validate a 2D transform.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the matrix represented by `x`; `n`
        number of columns in the matrix.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the matrix represented by the
        function's output; `no` number of columns in the matrix.

    """

    _levels('mn_tuple', (
        _generic(None, 'explicit collection', len_=2),
        _numeric(None, 'integer', range_='[1;inf)')))
    m, n = mn_tuple
    _numeric('x', ('integer', 'floating', 'complex'), shape=(m * n, 1))

    if overcomplete_mn_tuple is not None:
        _generic('overcomplete_mn_tuple', 'explicit collection', len_=2)
        _numeric(('overcomplete_mn_tuple', 0), 'integer',
                 range_='[{};inf)'.format(m))
        _numeric(('overcomplete_mn_tuple', 1), 'integer',
                 range_='[{};inf)'.format(n))


def _validate_transform_bwd(x, mn_tuple, overcomplete_mn_tuple=None):
    """
    Validate a 2D transform.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    mn_tuple : tuple of int
        `(m, n)` - `m` number of rows in the associated matrix, `n` number of
        columns in the associated matrix.
    overcomplete_mn_tuple : tuple of int, optional
        `(mo, no)` - `mo` number of rows in the associated matrix, `no` number
        of columns in the associated matrix.

    """

    _levels('mn_tuple', (
        _generic(None, 'explicit collection', len_=2),
        _numeric(None, 'integer', range_='[1;inf)')))
    m, n = mn_tuple

    if overcomplete_mn_tuple is None:
        shape = (m * n, 1)
    else:
        _generic('overcomplete_mn_tuple', 'explicit collection', len_=2)
        _numeric(('overcomplete_mn_tuple', 0), 'integer',
                 range_='[{};inf)'.format(m))
        _numeric(('overcomplete_mn_tuple', 1), 'integer',
                 range_='[{};inf)'.format(n))
        shape = (overcomplete_mn_tuple[0] * overcomplete_mn_tuple[1], 1)

    _numeric('x', ('integer', 'floating', 'complex'), shape=shape)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Package providing a toolbox for compressed sensing for atomic force microscopy.

Routine listings
----------------
afm
    Subpackage providing atomic force miscroscopy specific functionality.
cs
    Subpackage providing generic compressed sensing functionality.
imaging
    Subpackage providing generic imaging functionality.
tests
    Subpackage providing unittesting of the other subpackages.
utils
    Subpackage providing support functionality for the other subpackages.

Notes
-----
See the README file for additional information.

"""

from magni import afm
from magni import cs
from magni import imaging
from magni import reproducibility
from magni import utils

__version__ = '1.6.0'



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the core Iterative Thresholding (IT) algorithm.

Routine listings
----------------
run(y, A)
    Run the IT reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.it._config : Configuration options.
magni.cs.reconstruction.it._step_size : Available step-size strategies.
magni.cs.reconstruction.it._stop_criterion : Available stop criteria.
magni.cs.reconstruction.it._threshold : Available threshold estimators.
magni.cs.reconstruction.it._threshold_operators : Available threshold operators

Notes
-----
The default configuration of the IT algorithm provides the Iterative Hard
Thresholding (IHT) algorithm [1]_ using the False Alarm Rate (FAR) heuristic
from [2]_. By default a residual measurements ratio step criterion is used. The
algorithm may also be configured to act as the Iterative Soft Thresholding
(IST) [3]_ algorithm with the soft threshold from [4]_.

References
----------
.. [1] T. Blumensath and M.E. Davies, "Iterative Thresholding for Sparse
   Approximations", *Journal of Fourier Analysis and Applications*, vol. 14,
   pp. 629-654, Sep. 2008.
.. [2] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
   Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in Signal
   Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.
.. [3] I. Daubechies, M. Defrise, and C. D. Mol, "An Iterative Thresholding
   Algorithm for Linear Inverse Problems with a Sparsity Constraint",
   *Communications on Pure and Applied Mathematics*, vol. 57, no. 11,
   pp. 1413-1457, Nov. 2004.
.. [4] D.L. Donoho, "De-Noising by Soft-Thresholding", *IEEE Transactions on
   Information Theory*, vol. 41, no. 3, pp. 613-627, May. 1995.

"""

from __future__ import division

import numpy as np

from magni.cs.reconstruction.it import config as _conf
from magni.cs.reconstruction.it import _step_size
from magni.cs.reconstruction.it import _stop_criterion
from magni.cs.reconstruction.it import _threshold
from magni.cs.reconstruction.it import _threshold_operators
from magni.imaging.evaluation import calculate_mse as _calculate_mse
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def run(y, A):
    """
    Run the IT reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.
    history : dict, optional
        The dictionary of various measures tracked in the IT iterations.

    See Also
    --------
    magni.cs.reconstruction.it._config : Configuration options.
    magni.cs.reconstruction.it._step_size : Step-size calculation.
    magni.cs.reconstruction.it._stop_criterion : Stop criterion calculation.
    magni.cs.reconstruction.it._threshold : Threshold calculation.
    magni.cs.reconstruction.it._threshold_operators : Thresholding operators.

    Notes
    -----
    Optionally, the algorithm may be configured to save and return the
    iteration history along with the reconstruction result. The returned
    history contains the following:

    * alpha : Reconstruction coefficient estimates
    * MSE : Solution mean squared error (if the true solution is known).
    * stop_criterion : The currently used stop criterion.
    * stop_criterion_value : The value of the stop criterion.
    * stop_iteration : The iteration at which the algorithm stopped.
    * stop_reason : The reason for termination of the algorithm.

    Examples
    --------
    For example, recovering a vector from random measurements using IHT with
    the FAR heuristic

    >>> import numpy as np, magni
    >>> from magni.cs.reconstruction.it import run
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(90) * np.random.randn(90, 200)
    >>> alpha = np.zeros((200, 1))
    >>> alpha[:10] = 1
    >>> y = A.dot(alpha)
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:12]
    array([[ 0.99748945],
           [ 1.00082074],
           [ 0.99726507],
           [ 0.99987834],
           [ 1.00025857],
           [ 1.00003266],
           [ 1.00021666],
           [ 0.99838884],
           [ 1.00018356],
           [ 0.99859105],
           [ 0.        ],
           [ 0.        ]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10

    Or recover the same vector as above using IST with a fixed number of
    non-thresholded coefficients

    >>> it_config = {'threshold_operator': 'soft', 'threshold': 'fixed',
    ... 'threshold_fixed': 10}
    >>> magni.cs.reconstruction.it.config.update(it_config)
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:12]
    array([[ 0.99822443],
           [ 0.99888724],
           [ 0.9982493 ],
           [ 0.99928642],
           [ 0.99964131],
           [ 0.99947346],
           [ 0.9992809 ],
           [ 0.99872624],
           [ 0.99916842],
           [ 0.99903033],
           [ 0.        ],
           [ 0.        ]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10

    Or use a weighted IHT method with a fixed number of non-thresholded
    coefficietns to recover the above signal

    >>> weights = np.linspace(1.0, 0.9, 200).reshape(200, 1)
    >>> it_config = {'threshold_operator': 'weighted_hard',
    ... 'threshold_weights': weights}
    >>> magni.cs.reconstruction.it.config.update(it_config)
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:12]
    array([[ 0.99853888],
           [ 0.99886104],
           [ 0.99843149],
           [ 1.0000333 ],
           [ 1.00060273],
           [ 1.00035539],
           [ 0.99966219],
           [ 0.99912209],
           [ 0.99961541],
           [ 0.99952029],
           [ 0.        ],
           [ 0.        ]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10
    >>> magni.cs.reconstruction.it.config.reset()

    """

    @_decorate_validation
    def validate_input():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('A', ('integer', 'floating', 'complex'),
                 shape=(y.shape[0], -1))

    @_decorate_validation
    def validate_output():
        _numeric('alpha', ('integer', 'floating', 'complex'),
                 shape=(A.shape[1], 1), precision=convert(0).nbytes * 8)
        _generic('history', 'mapping',
                 keys_in=('alpha', 'MSE', 'stop_criterion',
                          'stop_criterion_value', 'stop_iteration',
                          'stop_reason'))

    validate_input()

    # Configured setup
    param = dict(_conf.items())
    convert = param['precision_float']
    iterations = param['iterations']
    report_history = param['report_history']
    tolerance = param['tolerance']
    true_solution = param['true_solution']
    threshold_weights = param['threshold_weights']
    kappa = param['kappa_fixed']
    threshold_alpha = _threshold_operators.get_function_handle(
        param['threshold_operator'])

    # Initialisation based on configuration
    if param['warm_start'] is not None:
        alpha = convert(param['warm_start'])
    else:
        alpha = np.zeros((A.shape[1], 1), dtype=convert)

    run.calculate_step_size = _step_size.get_function_handle(
        param['kappa'], locals())
    run.calculate_threshold = _threshold.get_function_handle(
        param['threshold'], locals())
    run.stop_criterion = _stop_criterion.get_function_handle(
        param['stop_criterion'], locals())

    history = {'alpha': [alpha],
               'MSE': [np.nan],
               'stop_criterion': param['stop_criterion'].upper(),
               'stop_criterion_value': [np.nan],
               'stop_iteration': 0,
               'stop_reason': 'MAX_ITERATIONS'}
    r = y.copy()

    # IT iterations
    for it in range(param['iterations']):
        alpha_prev = alpha
        c = A.conj().T.dot(r)

        # Step-size (relaxation parameter)
        kappa = run.calculate_step_size(locals())
        alpha = alpha + kappa * c

        # Threshold
        threshold = run.calculate_threshold(locals())
        threshold_alpha(locals())

        # Residual
        r = y - A.dot(alpha)

        # Stop criterion
        stop, stop_criterion_value = run.stop_criterion(locals())

        # History reporting
        if report_history:
            history['alpha'].append(alpha)
            history['stop_criterion_value'].append(stop_criterion_value)
            history['stop_iteration'] = it
            if true_solution is not None:
                history['MSE'].append(_calculate_mse(true_solution, alpha))

        if stop:
            history['stop_reason'] = param['stop_criterion'].upper()
            break

    validate_output()

    if report_history:
        return alpha, history
    else:
        return alpha



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
construct_measurement_matrix(coords, h, w)
    Function for constructing a measurement matrix.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements._util import unique_pixels as _unique_pixels
from magni.utils.matrices import Matrix as _Matrix
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['construct_measurement_matrix']


def construct_measurement_matrix(coords, h, w):
    """
    Construct a measurement matrix extracting the specified measurements.

    Parameters
    ----------
    coords : ndarray
        The `k` floating point coordinates arranged into a 2D array where each
        row is a coordinate pair (x, y), such that `coords` has size `k` x 2.
    h : int
        The height of the image measured in pixels.
    w : int
        The width of the image measured in pixels.

    Returns
    -------
    Phi : magni.utils.matrices.Matrix
        The constructed measurement matrix.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Notes
    -----
    The function constructs two functions: one for extracting pixels at the
    coordinates specified and one for the transposed operation. These functions
    are then wrapped by a matrix emulator which is returned.

    Examples
    --------
    Create a dummy 5 by 5 pixel image and an example sampling pattern:

    >>> import numpy as np, magni
    >>> img = np.arange(25, dtype=np.float).reshape(5, 5)
    >>> vec = magni.imaging.mat2vec(img)
    >>> coords = magni.imaging.measurements.uniform_line_sample_image(
    ...              5, 5, 16., 17)

    Sample the image in the ordinary way:

    >>> unique = magni.imaging.measurements.unique_pixels(coords)
    >>> samples_normal = img[unique[:, 1], unique[:, 0]]
    >>> samples_normal = samples_normal.reshape((len(unique), 1))

    Sample the image using the present function:

    >>> from magni.imaging.measurements import construct_measurement_matrix
    >>> matrix = construct_measurement_matrix(coords, *img.shape)
    >>> samples_matrix = matrix.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(samples_matrix, samples_normal)
    True

    """

    @_decorate_validation
    def validate_input():
        _numeric('coords', ('integer', 'floating'), shape=(-1, 2))
        _numeric('h', 'integer', range_='[1;inf)')
        _numeric('w', 'integer', range_='[1;inf)')
        _numeric('coords[:, 0]', ('integer', 'floating'),
                 range_='[0;{}]'.format(w), shape=(-1,), var=coords[:, 0])
        _numeric('coords[:, 1]', ('integer', 'floating'),
                 range_='[0;{}]'.format(h), shape=(-1,), var=coords[:, 1])

    validate_input()

    coords = _unique_pixels(coords)
    mask = coords[:, 0] * w + coords[:, 1]

    def measure(vec):
        return vec[mask]

    def measure_T(vec):
        output = np.zeros((h * w, 1), dtype=vec.dtype)
        output[mask] = vec
        return output

    return _Matrix(measure, measure_T, [], (len(mask), h * w))



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

A script that prints a dependency report for Magni

More about comparing versions of Python packages:
http://www.python.org/dev/peps/pep-0345/
http://www.python.org/dev/peps/pep-0386/
https://www.python.org/dev/peps/pep-0440/

"""

from __future__ import division
from pkg_resources import parse_version
import importlib
from itertools import chain
import platform
import re
import sys


# Package names vs import names
pac_names = {'python': 'Python',
             'numpy': 'NumPy',
             'scipy': 'SciPy',
             'tables': 'PyTables',
             'matplotlib': 'Matplotlib',
             'mkl': 'MKL',
             'sphinx': 'Sphinx',
             'IPython': 'IPython',
             'pyflakes': 'Pyflakes',
             'pep8': 'PEP8',
             'radon': 'Radon',
             'nose': 'Nose',
             'coverage': 'Coverage',
             'PIL': 'PIL',
             'bottleneck': 'Bottleneck',
             'conda': 'Conda',
             'setuptools': 'Setuptools',
             'sklearn': 'scikit-learn'}


# Minimum version requirements
python2_min_ver = '2.7'
python3_min_ver = '3.3'
python3_first_break_ver = '3.5'  # First version following minimum version for
# which some third party dependency requires a newer version than specified
# below.

deps = {'numpy': '1.8',
        'scipy': '0.14',
        'tables': '3.1',
        'matplotlib': '1.3',
        'setuptools': '11.3'}

opt_deps = {'mkl': '11.1',
            'sphinx': '1.3.1',
            'IPython': '2.1',
            'pyflakes': '0.8',
            'pep8': '1.5',
            'radon': '1.2',
            'nose': '1.3',
            'coverage': '3.7',
            'PIL': '1.1.7',
            'bottleneck': '1.0.0',
            'sklearn': '0.15.1',
            'conda': '3.7.0'}

ver_broken_opt_deps = {}

status = {}


# Python requirements
py_ver = platform.python_version()
if py_ver[0] == '2':
    if parse_version(py_ver) < parse_version(python2_min_ver):
        status['python'] = ('FAIL: Python 2 version nedded ' +
                            python2_min_ver + '; Installed version is ' +
                            py_ver)
    else:
        status['python'] = 'OK: Python 2 at version ' + py_ver

elif py_ver[0] == '3':
    if parse_version(py_ver) < parse_version(python3_min_ver):
        status['python'] = ('FAIL: Python 3 version needed ' +
                            python3_min_ver + '; Installed version is ' +
                            py_ver)
    else:
        status['python'] = 'OK: Python 3 at version ' + py_ver
        if parse_version(py_ver) >= parse_version(python3_first_break_ver):
            status['python'] = (
                'WARN: Python 3 at version ' + py_ver +
                ' (this version may require newer versions of some of the ' +
                'third party libraries)')

else:
    status['python'] = 'FAIL: Python version is ' + py_ver + '... What???'


# Third party dependencies
all_deps = dict(chain(deps.items(), opt_deps.items()))
for pac in all_deps:
    try:
        p = importlib.import_module(pac)

        # Fix packages that do not have a __version__ magic
        if pac == 'mkl':
            p.__version__ = re.search('\d+\.\d+\.\d+',
                                      p.get_version_string()).group()
        elif pac == 'PIL':
            import PIL.Image
            p.__version__ = PIL.Image.VERSION

        if parse_version(p.__version__) < parse_version(all_deps[pac]):
            status[pac] = ('WARN: Tested on version >= ' + all_deps[pac] +
                           '; Installed version is ' + p.__version__)
        else:
            status[pac] = 'OK: Installed version is ' + p.__version__

    except ImportError:
        status[pac] = 'FAIL: Unable to import'


# Version broken optional dependencies:
for brok, brok_ver in ver_broken_opt_deps.items():
    try:
        b = importlib.import_module(brok)
        if parse_version(b.__version__) < parse_version(brok_ver):
            status[brok] = ('WARN: Tested on version >= ' + brok_ver +
                            '; Installed version is ' + b.__version__)
        else:
            status[brok] = 'OK: Installed version is ' + b.__version__
    except AttributeError:
        status[brok] = ('WARN: Unknown version. Please ensure version >= ' +
                        ver_broken_opt_deps[brok])
    except ImportError:
        status[brok] = 'FAIL: Unable to import'


# Pretty print formatting function
def _print_dep_report(dep, status):
    """
    Pretty print a report about a single Magni dependency.

    Parameters
    ----------
    dep : str
        The key in the `status` dict correspoding to the relevant dependency.
    status : dict
        The dictionary holding the dependency status to be printed.

    """

    print ('{0:>14}  :  {1:<15}'.format(pac_names[dep], status[dep]))


# Print the full dependency report
print('\nMagni dependency report')
print('======================================================================')
print('\nPython installation:')

_print_dep_report('python', status)

print('\nThird party dependencies:')

for dep in sorted(deps):
    _print_dep_report(dep, status)

print('\nOptional third party dependencies:')

for opt in sorted(opt_deps):
    _print_dep_report(opt, status)

for opt in sorted(ver_broken_opt_deps):
    _print_dep_report(opt, status)

# Exit 1 in case of failures
for val in status.values():
    if 'FAIL' in val:
        sys.exit(1)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.reproducibility`.

**Testing Strategy**

Each annotation is a dict which must have specific keys. It is tested that
these keys are as expected.

Each chase is a string. It is tested that this string has certain properties.

"""

from __future__ import division
import inspect
import json
import os
import subprocess
import unittest

from pkg_resources import parse_version as _parse_version
import tables as tb

from magni.reproducibility import _annotation, _chase, io


class TestAnnotations(unittest.TestCase):
    """
    Test of annotations.

    """

    def setUp(self):
        self.ref_conda_410_info_keys = [
            'channels', 'conda_version', 'config_file', 'default_prefix',
            'env_export', 'envs_dirs', 'is_foreign_system', 'linked_modules',
            'modules_info', 'package_cache', 'platform', 'root_prefix',
            'status']
        self.ref_conda_420_info_keys = [
            'channels', 'conda_is_private', 'conda_version', 'config_file',
            'default_prefix', 'env_export', 'envs_dirs', 'linked_modules',
            'modules_info', 'offline_mode', 'package_cache', 'platform',
            'python_version', 'root_prefix', 'status']
        self.ref_datetime_keys = ['pretty_utc', 'status', 'today', 'utcnow']
        self.ref_git_revision_ok_keys = ['branch', 'remote', 'status', 'tag']
        self.ref_git_revision_nok_keys = ['output', 'returncode', 'status']
        self.ref_magni_config_keys = ['magni.afm.config',
                                      'magni.cs.phase_transition.config',
                                      'magni.cs.reconstruction.amp.config',
                                      'magni.cs.reconstruction.gamp.config',
                                      'magni.cs.reconstruction.iht.config',
                                      'magni.cs.reconstruction.it.config',
                                      'magni.cs.reconstruction.sl0.config',
                                      'magni.utils.multiprocessing.config',
                                      'status']
        self.ref_magni_info_keys = ['help_magni']
        self.ref_platform_info_keys = ['libc', 'linux', 'mac_os', 'machine',
                                       'node', 'processor', 'python',
                                       'release', 'status', 'system',
                                       'version', 'win32']
        self.file_ = 'file_hashes_test.txt'
        self.true_file_hashes = {
            'md5sum': '7af13f5773a4f06d1307cb2aeb9e08b8',
            'sha256sum': ('011e092f282c1dbba5fc3d58cff32e20' +
                          '753f6749f24bc58bd39da24984ac3391')}
        self.assertFalse(os.path.exists(self.file_))
        with open(self.file_, mode='w') as file_:
            file_.write('Some dummy text')

    def tearDown(self):
        if os.path.exists(self.file_):
            os.remove(self.file_)

    def test_get_conda_info(self):
        conda_info = _annotation.get_conda_info()
        conda_info_keys = sorted(list(conda_info.keys()))

        self.assertTrue('status' in conda_info_keys)

        if 'Failed' not in conda_info['status']:
            import conda
            if _parse_version(conda.__version__) >= _parse_version('4.2.0'):
                self.assertListEqual(
                    self.ref_conda_420_info_keys, conda_info_keys)
            else:
                self.assertListEqual(
                    self.ref_conda_410_info_keys, conda_info_keys)
            self.assertTrue(conda_info['status'] == 'Succeeded')

    def test_datetime(self):
        datetime_ = _annotation.get_datetime()
        datetime_keys = sorted(list(datetime_.keys()))

        self.assertListEqual(self.ref_datetime_keys, datetime_keys)
        self.assertTrue(datetime_['status'] == 'Succeeded')

    def test_file_hashes(self):
        file_hashes = _annotation.get_file_hashes(self.file_, blocksize=1)
        self.assertEqual(self.true_file_hashes, file_hashes)

    def test_get_git_revision(self):
        try:
            subprocess.check_output(['git', '--version'])
        except subprocess.CalledProcessError as e:
            pass
        except OSError as e:
            pass
        else:
            git_revision = _annotation.get_git_revision()
            git_revision_keys = sorted(list(git_revision.keys()))
            git_revision_here = _annotation.get_git_revision('.')
            git_revision_here_keys = sorted(list(git_revision_here.keys()))

            self.assertTrue('status' in git_revision_keys)
            self.assertTrue('status' in git_revision_here_keys)

            if git_revision['status'] == 'Succeeded':
                self.assertListEqual(self.ref_git_revision_ok_keys,
                                     git_revision_keys)
            else:
                self.assertListEqual(self.ref_git_revision_nok_keys,
                                     git_revision_keys)

        with self.assertRaises(OSError):
            _annotation.get_git_revision('this/should/not/be/a/directory')

    def test_get_magni_config(self):
        magni_config = _annotation.get_magni_config()
        magni_config_keys = sorted(list(magni_config.keys()))

        self.assertTrue('status' in magni_config_keys)
        self.assertTrue(magni_config['status'] == 'Succeeded')
        self.assertListEqual(self.ref_magni_config_keys, magni_config_keys)

    def test_get_magni_info(self):
        magni_info = _annotation.get_magni_info()
        magni_info_keys = sorted(list(magni_info.keys()))

        self.assertListEqual(self.ref_magni_info_keys, magni_info_keys)

    def test_get_platform_info(self):
        platform_info = _annotation.get_platform_info()
        platform_info_keys = sorted(list(platform_info.keys()))

        self.assertFalse('failed' in platform_info['status'])
        self.assertListEqual(self.ref_platform_info_keys, platform_info_keys)


class TestChases(unittest.TestCase):
    """
    Test of chases.

    """

    def setUp(self):
        self.cur_dir = os.getcwd()
        try:
            # chdir in run_tests.py workaround
            os.chdir(inspect.stack()[-2][0].f_locals['cur_dir'])
        except KeyError:
            pass

    def tearDown(self):
        os.chdir(self.cur_dir)

    def test_get_main_file_name(self):
        main_file_name = _chase.get_main_file_name()
        self.assertIsInstance(main_file_name, str)
        self.assertNotEqual(main_file_name, '')
        self.assertNotIn('Failed', main_file_name)

    def test_get_main_file_source(self):
        main_file_source = _chase.get_main_file_source()
        self.assertIsInstance(main_file_source, str)
        self.assertNotEqual(main_file_source, '')
        self.assertNotIn('Failed', main_file_source)

    def test_get_main_source(self):
        main_source = _chase.get_main_source()
        self.assertIsInstance(main_source, str)
        self.assertNotEqual(main_source, '')
        self.assertNotIn('Failed', main_source)

    def test_get_stack_trace(self):
        stack_trace = _chase.get_stack_trace()
        self.assertIsInstance(stack_trace, str)
        self.assertNotEqual(stack_trace, '')
        self.assertNotIn('Failed', stack_trace)


class TestIO(unittest.TestCase):
    """
    Tests of IO functions.

    """

    def setUp(self):
        self.h5_name = 'db.hdf5'
        self.assertFalse(os.path.exists(self.h5_name))
        self.ref_annotation_keys = ['conda_info', 'datetime', 'git_revision',
                                    'magni_config', 'magni_info',
                                    'platform_info']
        self.ref_chase_keys = ['main_file_name', 'main_file_source',
                               'main_source', 'stack_trace']

    def tearDown(self):
        if os.path.exists(self.h5_name):
            os.remove(self.h5_name)

    def test_annotations_roundtrip(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            io.annotate_database(h5_file)
            annotation_keys = sorted(io.read_annotations(h5_file).keys())
            self.assertEqual(annotation_keys, self.ref_annotation_keys)
            io.remove_annotations(h5_file)
            with self.assertRaises(tb.NoSuchNodeError):
                io.read_annotations(h5_file)

    def test_annotations_exceptions(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            # No annotations
            io.remove_annotations(h5_file)  # Pass if no annotations
            with self.assertRaises(tb.NoSuchNodeError):
                io.read_annotations(h5_file)

            # Non-compliant annotation
            annotations_group = h5_file.create_group('/', 'annotations')
            h5_file.create_array(annotations_group, 'fail_ann',
                                 obj='[fail)'.encode())
            h5_file.flush()
            with self.assertRaises(ValueError):
                io.read_annotations(h5_file)

            # Already annotated database
            with self.assertRaises(tb.NodeError):
                io.annotate_database(h5_file)

    def test_chases_roundtrip(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            io.chase_database(h5_file)
            chase_keys = sorted(io.read_chases(h5_file).keys())
            self.assertEqual(chase_keys, self.ref_chase_keys)
            io.remove_chases(h5_file)
            with self.assertRaises(tb.NoSuchNodeError):
                io.read_chases(h5_file)

    def test_chases_exceptions(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            # No chases
            io.remove_chases(h5_file)  # Pass if no chases
            with self.assertRaises(tb.NoSuchNodeError):
                io.read_chases(h5_file)

            # Non-complient chase
            chases_group = h5_file.create_group('/', 'chases')
            h5_file.create_array(chases_group, 'fail_chase',
                                 obj='[fail)'.encode())
            h5_file.flush()
            with self.assertRaises(ValueError):
                io.read_chases(h5_file)

            # Already chased database
            with self.assertRaises(tb.NodeError):
                io.chase_database(h5_file)

    def test_create_database(self):
        io.create_database(self.h5_name)
        with tb.File(self.h5_name, mode='a') as h5_file:
            annotation_keys = sorted(io.read_annotations(h5_file).keys())
            self.assertEqual(annotation_keys, self.ref_annotation_keys)
            chase_keys = sorted(io.read_chases(h5_file).keys())
            self.assertEqual(chase_keys, self.ref_chase_keys)

        # Overwrite test
        io.create_database(self.h5_name)
        with self.assertRaises(IOError):
            io.create_database(self.h5_name, overwrite=False)

    def test_write_custom_annotation(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            # Roundtrip
            io.write_custom_annotation(h5_file, 'test_ann', 'a test')

            self.assertEqual(
                'a test',
                json.loads(h5_file.root.annotations.test_ann.read().decode()))

            io.remove_annotations(h5_file)

            with self.assertRaises(tb.NoSuchNodeError):
                h5_file.root.annotations.test_ann.read()

            # Invalid annotation
            with self.assertRaises(TypeError):
                io.write_custom_annotation(h5_file, 'fail_ann', lambda x: x)

            # Annotations subgroup
            io.write_custom_annotation(h5_file, 'test_ann', 'a test',
                                       annotations_sub_group='test_group')

            self.assertEqual(
                'a test',
                json.loads(
                    h5_file.root.annotations.test_group.test_ann.read(
                        ).decode()
                ))

            io.remove_annotations(h5_file)

            with self.assertRaises(tb.NoSuchNodeError):
                h5_file.root.annotations.test_group.test_ann.read()

            # Writing to existing annotation
            io.write_custom_annotation(h5_file, 'test_ann', 'a test')
            with self.assertRaises(tb.NodeError):
                io.write_custom_annotation(h5_file, 'test_ann', 'a test')

    def test_recursive_annotation_read(self):
        with tb.File(self.h5_name, mode='a') as h5_file:
            io.annotate_database(h5_file)
            io.write_custom_annotation(
                h5_file, 'test_ann_1', 'a test 1',
                annotations_sub_group='zzz_test_group_1')
            io.write_custom_annotation(
                h5_file, 'test_ann_2', 'a test 2',
                annotations_sub_group='zzz_test_group_2')
            io.write_custom_annotation(
                h5_file, 'test_ann_11', 'a test 11',
                annotations_sub_group='zzz_test_group_1/zzz_test_group_11')

            annotations = io.read_annotations(h5_file)
            annotation_lvl_0_keys = sorted(annotations.keys())
            self.assertEqual(
                annotation_lvl_0_keys,
                self.ref_annotation_keys +
                ['zzz_test_group_1', 'zzz_test_group_2'])
            annotation_lvl_01_keys = sorted(annotations['zzz_test_group_1'])
            self.assertEqual(
                annotation_lvl_01_keys,
                ['test_ann_1', 'zzz_test_group_11'])
            annotation_lvl_02_keys = sorted(annotations['zzz_test_group_2'])
            self.assertEqual(annotation_lvl_02_keys, ['test_ann_2'])
            annotation_lvl_11_keys = sorted(
                annotations['zzz_test_group_1']['zzz_test_group_11'])
            self.assertEqual(annotation_lvl_11_keys, ['test_ann_11'])



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating the starting value of sigma used in
the Smoothed l0 algorithms.

Routine listings
----------------
calculate_using_fixed(var)
    Calculate the fixed sigma value.
calculate_using_reciprocal(var)
    Calculate a sigma value in a 'reciprocal' way.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division


def wrap_calculate_using_fixed(var):
    """
    Arguments wrapper for calculate_using_fixed.

    """

    convert = var['convert']
    sigma_start = convert(var['param']['sigma_start_fixed'])

    def calculate_using_fixed():
        """
        Calculate the fixed sigma value.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the sigma
            value.

        Returns
        -------
        sigma : float
            The sigma value to be used in the SL0 algorithm.

        """

        return sigma_start

    return calculate_using_fixed


def wrap_calculate_using_reciprocal(var):
    """
    Arguments wrapper for calculate_using_reciprocal.

    """

    convert = var['convert']
    delta = var['A'].shape[0] / var['A'].shape[1]
    factor = convert(var['param']['sigma_start_reciprocal'])

    def calculate_using_reciprocal():
        """
        Calculate a sigma value in a 'reciprocal' way.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the sigma value.

        Returns
        -------
        sigma : float
            The sigma value to be used in the SL0 algorithm.

        """

        return 1 / (factor * delta)

    return calculate_using_reciprocal


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the sigma method.

    Returns
    -------
    f_handle : function
        Handle to the calculation method defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of Smoothed l0 Norm (SL0).

The implementations provided are the original SL0 reconstruction algorithm and
a modified SL0 reconstruction algorithm. The algorithm used depends on the
'algorithm' configuration option: 'std' refers to the original algorithm while
'mod' (default) refers to the modified algorithm.

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
run(y, A)
    Run the specified SL0 reconstruction algorithm.

Notes
-----
See `_algorithm` for documentation of `run`.

Implementations of the original SL0 reconstruction algorithm [1]_ and a
modified Sl0 reconstruction algorithm [3]_ are available. It is also possible
to configure the subpackage to provide customised versions of the SL0
reconstruction algorithm. The projection algorithm [1]_ is used for small delta
(< 0.55) whereas the contraint elimination algorithm [2]_ is used for large
delta (>= 0.55) which merely affects the computation time.

References
----------
.. [1] H. Mohimani, M. Babaie-Zadeh, and C. Jutten, "A Fast Approach for
   Overcomplete Sparse Decomposition Based on Smoothed l0 Norm", *IEEE
   Transactions on Signal Processing*, vol. 57, no. 1, pp. 289-301, Jan. 2009.
.. [2] Z. Cui, H. Zhang, and W. Lu, "An Improved Smoothed l0-norm Algorithm
   Based on Multiparameter Approximation Function", *in 12th IEEE International
   Conference on Communication Technology (ICCT)*, Nanjing, China, Nov. 11-14,
   2011, pp. 942-945.
.. [3] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from magni.cs.reconstruction.sl0._config import configger as config
from magni.cs.reconstruction.sl0._algorithm import run



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing data container classes for .mi image files.

The classes of this module can be used either directly or indirectly through
the `magni.afm.io` subpackage by loading an .mi image file.

Routine listings
----------------
Buffer(magni.afm.types.BaseClass)
    Data class of the .mi image file buffers.
Image(magni.afm.types.File)
    Data class of the .mi image files.

See Also
--------
magni.afm.io : .mi file loading.

"""

from __future__ import division
from collections import OrderedDict as _OrderedDict
import types

from magni.afm.types import BaseClass as _BaseClass
from magni.afm.types import File as _File
from magni.imaging.preprocessing import detilt as _detilt
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


class Buffer(_BaseClass):
    """
    Data class of the .mi image file buffers.

    Parameters
    ----------
    attrs : dict
        The attributes of the buffer.
    data : numpy.ndarray
        The 2D data of the buffer.

    Attributes
    ----------
    apply_clipping : bool
        A flag indicating if clipping should be applied to the data.
    data : numpy.ndarray
        The 2D data of the buffer.

    See Also
    --------
    magni.utils.types.BaseClass : Superclass of the present class.

    Examples
    --------
    A subclass of the present class is implicitly instantiated when loading,
    for example, the .mi file provided with the package:

    >>> import os, magni
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     image = magni.afm.io.read_mi_file(path)
    ...     buffer_ = image.buffers[0]

    This buffer can have a number of attributes including 'bufferLabel':

    >>> if os.path.isfile(path):
    ...     print('{!r}'.format(buffer_.attrs['bufferLabel']))
    ... else:
    ...     print("'Topography'")
    'Topography'

    The primary purpose of this class is, however, to contain the 2D data of a
    buffer:

    >>> if os.path.isfile(path):
    ...     print('Buffer data shape: {!r}'.format(
    ...     tuple(int(s) for s in buffer_.data.shape)))
    ... else:
    ...     print('Buffer data shape: (256, 256)')
    Buffer data shape: (256, 256)

    """

    _params = _OrderedDict((('bufferLabel', str),
                            ('trace', bool),
                            ('bufferUnit', str),
                            ('bufferRange', float),
                            ('DisplayOffset', float),
                            ('DisplayRange', float),
                            ('filter', str)))

    def __init__(self, attrs, data):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=('bufferLabel',))
            _numeric('data', ('integer', 'floating'), shape=(-1, -1))

        _BaseClass.__init__(self, attrs)
        validate_input()

        self._apply_clipping = False
        self._data = data
        self._data.flags['WRITEABLE'] = False

    @property
    def apply_clipping(self):
        """
        Get the apply_clipping property of the buffer.

        The clipping is specified by the 'DisplayOffset', 'DisplayRange', and
        'filter' attributes of the buffer.

        Returns
        -------
        apply_clipping : bool
            A flag indicating if clipping should be applied to the data.

        """

        return self._apply_clipping

    @apply_clipping.setter
    def apply_clipping(self, value):
        """
        Set the apply_clipping property of the buffer.

        The clipping is specified by the 'DisplayOffset', 'DisplayRange', and
        'filter' attributes of the buffer.

        Parameters
        ----------
        value : bool
            The desired value of the property.

        """

        @_decorate_validation
        def validate_input():
            _numeric('value', 'boolean')

        validate_input()

        self._apply_clipping = value

    @property
    def data(self):
        """
        Get the data property of the buffer.

        If `apply_clipping` is False, the data is returned as-is. Otherwise,
        the filtering specified by the 'filter' attribute of the buffer and the
        clipping specified by the 'DisplayOffset' and 'DisplayRange' attributes
        of the buffer are applied to the data before it is returned.

        Returns
        -------
        data : numpy.ndarray
            The 2D data of the buffer.

        """

        data = self._data
        attrs = self.attrs

        if self.apply_clipping:
            try:
                filter_ = attrs['filter']
                offset = attrs['DisplayOffset']
                range_ = attrs['DisplayRange']
            except KeyError:
                msg = ("The value of >>self.attrs<<, {!r}, must have the keys "
                       "('filter', 'DisplayOffset', 'DisplayRange'). when "
                       "self.apply_filter is True.")
                raise KeyError(msg.format(attrs))

            if filter_ == 'None':
                data = data.copy()
            elif filter_ == '1st_order':
                data = _detilt(data, mode='line_flatten', degree=1)
            elif filter_ == '2nd_order':
                data = _detilt(data, mode='line_flatten', degree=2)
            elif filter_ == '3rd_order':
                data = _detilt(data, mode='line_flatten', degree=3)
            else:
                msg = ("The value of >>self.attrs['filter']<<, {!r}, must be "
                       "in ('None', '1st_order', '2nd_order', '3rd_order').")
                raise ValueError(msg.format(attrs['filter']))

            data[data < offset - 0.5 * range_] = offset - 0.5 * range_
            data[data > offset + 0.5 * range_] = offset + 0.5 * range_

        return data


class Image(_File):
    """
    Data class of the .mi image files.

    Parameters
    ----------
    attrs : dict
        The attributes of the image.
    buffers : list or tuple
        The buffers of the image.

    See Also
    --------
    magni.utils.types.File : Superclass of the present class.

    Examples
    --------
    The present class is implicitly instantiated when loading, for example, the
    .mi file provided with the package:

    >>> import os, magni
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     image = magni.afm.io.read_mi_file(path)

    This file has a number of buffers which each has the 'bufferLabel'
    attribute:

    >>> if os.path.isfile(path):
    ...     for buffer_ in image.buffers[::2][:3]:
    ...         label = buffer_.attrs['bufferLabel']
    ...         print("Buffer with 'bufferLabel': {!r}".format(label))
    ... else:
    ...     for label in ('Topography', 'Deflection', 'Friction'):
    ...         print("Buffer with 'bufferLabel': {!r}".format(label))
    Buffer with 'bufferLabel': 'Topography'
    Buffer with 'bufferLabel': 'Deflection'
    Buffer with 'bufferLabel': 'Friction'

    """

    _params = _OrderedDict((('mode', str),
                            ('xPixels', int),
                            ('yPixels', int),
                            ('scanUp', bool),
                            ('scanSpeed', float),
                            ('acMac', bool),
                            ('acACMode', bool)))

    def __init__(self, attrs, buffers):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=('xPixels', 'yPixels'))
            _levels('buffers', (_generic(None, 'explicit collection'),
                                _generic(None, Buffer)))

            attrs = self.attrs

            for i, buffer_ in enumerate(buffers):
                if buffer_.attrs['bufferLabel'][:9] != 'Thumbnail':
                    _numeric('buffers[{}].data'.format(i),
                             ('integer', 'floating'),
                             shape=(attrs['yPixels'], attrs['xPixels']),
                             var=buffer_.data)

        _File.__init__(self, attrs, buffers)
        validate_input()



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

A setup.py script for Magni.

More about python packaging:
https://python-packaging-user-guide.readthedocs.org
https://pythonhosted.org/setuptools/setuptools.html#developer-s-guide
https://github.com/pypa/sampleproject

"""

import io
import re

from setuptools import setup, find_packages


def get_version(file_path):
    """
    Return the project version tag.

    Parameters
    ----------
    file_path : str
        The path to the project __init__.py file.

    Returns
    -------
    version : str
        The project version tag.

    """

    with open(file_path, mode='r') as init_file:
        init_cont = init_file.read()

    # PEP 440: [N:]N(.N)*[{a|b|c|rc}N][.postN][.devN]
    ver_spec = (r'__version__ = ["\'](\d+:)?\d+(\.\d+){1,2}' +
                r'((a|b|c|rc)\d+)?(\.post\d+)?(\.dev\d+)?["\']')
    ver_match = re.search(ver_spec, init_cont)

    if not ver_match:
        raise RuntimeError('Unable to find (valid) version specifier in {0!r}'
                           .format(file_path))

    return ver_match.group(0)[15:-1]


with io.open('README.rst', mode='r', encoding='utf-8') as desc_file:
    long_description = desc_file.read()


setup(
    name='magni',
    version=get_version('magni/__init__.py'),
    description=('A Python Package for Compressive Sampling ' +
                 'and Reconstruction of Atomic Force Microscopy Images'),
    long_description=long_description,
    url='https://github.com/SIP-AAU/Magni',
    download_url='https://pypi.python.org/pypi/magni',
    author='Magni Developers',
    author_email='magni@es.aau.dk',

    license='BSD 2-Clause',

    # Classifiers
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization'],

    # Keywords
    keywords=('Atomic Force Microscopy; Compressive Sensing; Python; ' +
              'Image Reconstruction; Reproducible Research'),

    # Project packages
    packages=find_packages(),

    # Required runtime dependencies
    install_requires=['numpy>=1.8',
                      'scipy>=0.13',
                      'tables>=3.1',
                      'matplotlib>=1.3',
                      'setuptools>=11.3'],

    # Data files that are part of package
    package_data={},
    data_files=[('', ['LICENSE.rst', 'CHANGELOG.rst'])]

)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
uniform_line_sample_image(h, w, scan_length, num_points)
    Function for uniform line sampling an image.
uniform_line_sample_surface(l, w, speed, sample_rate, time)
    Function for uniform line sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['uniform_line_sample_image', 'uniform_line_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def uniform_line_sample_image(h, w, scan_length, num_points):
    """
    Sample an image using a set of uniformly distributed straight lines.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a pattern based on a set of uniformly distributed
    straight lines are determined. The `scan_length` determines the length of
    the path scanned whereas `num_points` indicates the number of samples taken
    on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import uniform_line_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> uniform_line_sample_image(h, w, scan_length, num_points)
    array([[ 0.5       ,  0.5       ],
           [ 4.59090909,  0.5       ],
           [ 8.68181818,  0.5       ],
           [ 9.22727273,  3.5       ],
           [ 5.13636364,  3.5       ],
           [ 1.04545455,  3.5       ],
           [ 1.04545455,  6.5       ],
           [ 5.13636364,  6.5       ],
           [ 9.22727273,  6.5       ],
           [ 8.68181818,  9.5       ],
           [ 4.59090909,  9.5       ],
           [ 0.5       ,  9.5       ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))

    validate_input()

    coords = uniform_line_sample_surface(float(h - 1), float(w - 1),
                                         scan_length, float(num_points - 1),
                                         1.0)

    coords = coords + 0.5

    return coords


def uniform_line_sample_surface(l, w, speed, sample_rate, time):
    """
    Sample aa surface area using a set of uniformly distributed straight lines.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a pattern based on a set of uniformly distributed
    straight lines are determined.  The scanned path is determined from the
    probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `l` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import uniform_line_sample_surface
    >>> l = 2e-6
    >>> w = 2e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> uniform_line_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.        ,  0.        ],
           [ 0.00000067,  0.        ],
           [ 0.00000133,  0.        ],
           [ 0.000002  ,  0.        ],
           [ 0.000002  ,  0.00000067],
           [ 0.00000167,  0.000001  ],
           [ 0.000001  ,  0.000001  ],
           [ 0.00000033,  0.000001  ],
           [ 0.        ,  0.00000133],
           [ 0.        ,  0.000002  ],
           [ 0.00000067,  0.000002  ],
           [ 0.00000133,  0.000002  ],
           [ 0.000002  ,  0.000002  ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))

    validate_input()

    num_lines = int(np.floor((speed * time - l) / w))

    # We should always at least partially scan top and bottom lines.
    if num_lines < 2:
        num_lines = 2

    coords = np.zeros((2 * num_lines, 2))
    coords[1::4, 0] = coords[2::4, 0] = w
    coords[0::2, 1] = coords[1::2, 1] = np.linspace(0, l, num_lines)

    return _util.sample_lines(coords, speed, sample_rate, time)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing miscellaneous utility functions for the different
implementations of Iterative Thresholding (IT).

Routine listings
----------------
get_methods(module)
    Extract relevant methods from module.
def _get_operators(module)
    Extract relevant operators from `module`.

"""

from __future__ import division

from magni.cs.reconstruction.it import _stop_criterion
from magni.cs.reconstruction.it import _step_size
from magni.cs.reconstruction.it import _threshold
from magni.cs.reconstruction.it import _threshold_operators


def _get_methods(module):
    """
    Extract relevant methods from `module`.

    Parameters
    ----------
    module : str
        The name of the module from which the methods should be extracted.

    Returns
    -------
    methods : tuple
        The names of the relevant methods from `module`.

    Notes
    -----
    Looks for functions with names starting with 'wrap_calculate_using\_' in
    the module: _"module".

    """

    method_candidates = eval('_' + module).__dict__.keys()
    methods = [candidate[21:] for candidate in method_candidates
               if candidate[:21] == 'wrap_calculate_using_']

    return tuple(methods)


def _get_operators(module):
    """
    Extract relevant operators from `module`.

    Parameters
    ----------
    module : str
        The name of the module from which the operators should be extracted.

    Returns
    -------
    methods : tuple
        The names of the relevant operators from `module`.

    Notes
    -----
    Looks for functions with names starting with 'threshold\_' in the
    module _"module".

    """

    operator_candidates = eval('_' + module).__dict__.keys()
    operators = [candidate[10:] for candidate in operator_candidates
                 if candidate[:10] == 'threshold_']

    return tuple(operators)



r"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.sl0`
subpackage.

See also
--------
magni.cs.reconstruction._config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

epsilon : float
    The precision parameter used in centering (the default is 0.01).
L : {'geometric', 'fixed'}
    The method for selecting the maximum number of gradient descent iterations
    for each sigma.
L_fixed : int
    The value used for L if using the 'fixed' method (the default is 2.0).
L_geometric_ratio : float
    The common ratio used for the L update if using the 'geometric' method (the
    default is 2.0).
L_geometric_start : float
    The starting value used for the L if using the 'geometric' method (the
    default is 2.0).
mu : {'step', 'fixed'}
    The method for selecting the relative step-size used in gradient descent
    iteration.
mu_fixed : float
    The value used for mu if using the 'fixed' method (the default is 1.0).
mu_step_end : float
    The value used for mu for the last iterations if using the 'step' method
    (the default is 1.5).
mu_step_iteration : int
    The iteration where the value used for mu changes if using the 'step'
    method
mu_step_start : float
    The value used for mu for the first iterations if using the 'step' method
    (the default is 0.001).
precision_float : {np.float, np.float16, np.float32, np.float64, np.float128}
    The floating point precision used for the computations (the default is
    np.float64).
sigma_geometric : float
    The common ratio used for the sigma update (the default is 0.7).
sigma_start : {'reciprocal', 'fixed'}
    The method for selecting the starting value of sigma.
sigma_start_fixed : float
    The fixed factor multiplied onto the maximum coefficient of a least squares
    solution to obtain the value if using the 'fixed' method (the
    default is 2.0).
sigma_start_reciprocal : float
    The constant in the factor :math:`\frac{1}{constant \cdot \delta}`
    multiplied onto the maximum coefficient of a least squares solution to
    obtain the value if using the 'reciprocal' method (the default is 2.75).
sigma_stop_fixed : float
    The minimum value of std. dev. in Gaussian l0 approx (the default 0.01).

"""

from __future__ import division

import numpy as np

from magni.cs.reconstruction._config import Configger as _Configger
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


configger = _Configger(
    {'epsilon': 1e-2,
     'L': 'geometric',
     'L_fixed': 2,
     'L_geometric_ratio': 2.0,
     'L_geometric_start': 2.0,
     'mu': 'step',
     'mu_fixed': 1.0,
     'mu_step_end': 1.5,
     'mu_step_iteration': 4,
     'mu_step_start': 0.001,
     'precision_float': np.float64,
     'sigma_geometric': 0.7,
     'sigma_start': 'reciprocal',
     'sigma_start_fixed': 2.0,
     'sigma_start_reciprocal': 2.75,
     'sigma_stop_fixed': 0.01},
    {'epsilon': _numeric(None, 'floating', range_='(0;inf)'),
     'L': _generic(None, 'string', value_in=('fixed', 'geometric')),
     'L_fixed': _numeric(None, 'integer', range_='[1;inf)'),
     'L_geometric_ratio': _numeric(None, 'floating', range_='(0;inf)'),
     'L_geometric_start': _numeric(None, 'floating', range_='[1;inf)'),
     'mu': _generic(None, 'string', value_in=('fixed', 'step')),
     'mu_fixed': _numeric(None, 'floating', range_='(0;inf)'),
     'mu_step_end': _numeric(None, 'floating', range_='(0;inf)'),
     'mu_step_iteration': _numeric(None, 'integer', range_='(0;inf)'),
     'mu_step_start': _numeric(None, 'floating', range_='(0;inf)'),
     'precision_float': _generic(None, type, value_in=(
         np.float,
         getattr(np, 'float16', np.float_),
         getattr(np, 'float32', np.float_),
         getattr(np, 'float64', np.float_),
         getattr(np, 'float128', np.float_))),
     'sigma_geometric': _numeric(None, 'floating', range_='(0;1)'),
     'sigma_start': _generic(None, 'string', value_in=('fixed', 'reciprocal')),
     'sigma_start_fixed': _numeric(None, 'floating', range_='(0;inf)'),
     'sigma_start_reciprocal': _numeric(None, 'floating', range_='(0;inf)'),
     'sigma_stop_fixed': _numeric(None, 'floating', range_='(0;inf)')})



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input/output functionality for stored phase transitions.

Routine listings
----------------
load_phase_transition(path, label='default')
    Load the coordinates of a phase transition from a HDF5 file.

"""

from __future__ import division

import numpy as np

from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def load_phase_transition(path, label='default', delta=None):
    """
    Load the coordinates of a phase transition from a HDF5 file.

    This function is used to load the phase transition from the output file
    generated by `magni.cs.phase_transition.determine`.

    Parameters
    ----------
    path : str
        The path of the HDF5 file where the phase transition is stored.
    label : str
        The label assigned to the phase transition (the default is 'default').
    delta : ndarray
        The undersampling ratio values to use (the default is None which
        implies that a uniform spacing of values in the interval (0,1] is
        assumed.)

    Returns
    -------
    delta : ndarray
        The delta values of the phase transition points.
    rho : ndarray
        The rho values of the phase transition points.

    See Also
    --------
    magni.cs.phase_transition.determine : Phase transition determination.
    magni.cs.phase_transition.plotting : Phase transition plotting.

    Examples
    --------
    An example of how to use this function is provided in the `examples` folder
    in the `cs-phase_transition.ipynb` ipython notebook file.

    """

    @_decorate_validation
    def validate_input():
        _generic('path', 'string')
        _generic('label', 'string')
        _numeric('delta', ('floating', 'integer'), range_='[0;1]', shape=(-1,),
                 ignore_none=True)

    @_decorate_validation
    def validate_output():
        _numeric('rho', ('floating', 'integer'), range_='[0;1]', shape=(-1,))
        _numeric('delta', ('floating', 'integer'), range_='[0;1]',
                 shape=rho.shape)

    validate_input()

    with _File(path, 'r') as f:
        rho = f.get_node('/' + label + '/phase_transition')[:]

    if delta is None:
        delta = np.linspace(0, 1, len(rho) + 1)[1:]

    validate_output()

    return (delta, rho)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input/output functions to databases containing results from
reproducible research.

Routine listings
----------------
annotate_database(h5file)
    Function for annotating an existing HDF5 database.
chase_database(h5file)
    Function for chasing an existing HDF5 database.
create_database(h5file)
    Function for creating a new annotated and chased HDF5 database.
read_annotations(h5file)
    Function for reading annotations in an HDF5 database.
read_chases(h5file)
    Function for reading chases in an HDF5 database.
remove_annotations(h5file)
    Function for removing annotations in an HDF5 database.
remove_chases(h5file)
    Function for removing chases in an HDF5 database.
write_custom_annotation(h5file, annotation_name, annotation_value,
    annotations_sub_group=None)
    Write a custom annotation to an HDF5 database.

See Also
--------
magni.reproducibility._annotation.get_conda_info : Conda annotation
magni.reproducibility._annotation.get_git_revision : Git annotation
magni.reproducibility._annotation.get_platform_info : Platform annotation
magni.reproducibility._annotation.get_datetime : Date and time annotation
magni.reproducibility._annotation.get_magni_config : Magni config annotation
magni.reproducibility._annotation.get_magni_info : Magni info annotation
magni.reproducibility._chase.get_main_file_name : Magni main file name chase
magni.reproducibility._chase.get_main_file_source : Magni source code chase
magni.reproducibility._chase.get_main_source : Magni main source code chase
magni.reproducibility._chase.get_stack_trace : Magni stack trace chase

"""

from __future__ import division
import json
import os

import tables

from magni.reproducibility import _annotation
from magni.reproducibility import _chase
from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def annotate_database(h5file):
    """
    Annotate an HDF5 database with information about Magni and the platform.

    The annotation consists of a group in the root of the `h5file` having nodes
    that each provide information about Magni or the platform on which this
    function is run.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database that should be annotated.

    See Also
    --------
    magni.reproducibility._annotation.get_conda_info : Conda annotation
    magni.reproducibility._annotation.get_git_revision : Git annotation
    magni.reproducibility._annotation.get_platform_info : Platform annotation
    magni.reproducibility._annotation.get_datetime : Date and time annotation
    magni.reproducibility._annotation.get_magni_config : Magni config
        annotation
    magni.reproducibility._annotation.get_magni_info : Magni info annotation

    Notes
    -----
    The annotations of the database includes the following:

    * conda_info - Information about Continuum Anacononda install
    * git_revision - Git revision and tag of Magni
    * platform_info - Information about the current platform (system)
    * datetime - The current date and time
    * magni_config - Infomation about the current configuration of Magni
    * magni_info - Information from `help(magni)`

    Examples
    --------
    Annotate the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import annotate_database
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...     annotate_database(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    annotations = {'conda_info': json.dumps(_annotation.get_conda_info()),
                   'git_revision': json.dumps(_annotation.get_git_revision()),
                   'platform_info': json.dumps(
                       _annotation.get_platform_info()),
                   'datetime': json.dumps(_annotation.get_datetime()),
                   'magni_config': json.dumps(_annotation.get_magni_config()),
                   'magni_info': json.dumps(_annotation.get_magni_info())}

    try:
        annotations_group = h5file.create_group('/', 'annotations')
        for annotation in annotations:
            h5file.create_array(annotations_group, annotation,
                                obj=annotations[annotation].encode())
        h5file.flush()
    except tables.NodeError:
        raise tables.NodeError('The database has already been annotated. ' +
                               'Remove the existing annotation prior to ' +
                               '(re)annotating the database.')


def chase_database(h5file):
    """
    Chase an HDF5 database to track information about stack and source code.

    The chase consist of a group in the root of the `h5file` having nodes that
    each profide information about the program execution that led to this chase
    of the database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database that should be chased.

    See Also
    --------
    magni.reproducibility._chase.get_main_file_name : Name of main file
    magni.reproducibility._chase.get_main_file_source : Main file source code
    magni.reproducibility._chase.get_main_source : Source code around main
    magni.reproducibility._chase.get_stack_trace : Complete stack trace

    Notes
    -----
    The chase include the following information:

    * main_file_name - Name of the main file/script that called this function
    * main_file_source - Full source code of the main file/script
    * main_source - Extract of main file source code that called this function
    * stack_trace - Complete stack trace up until the call to this function

    Examples
    --------
    Chase the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import chase_database
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...     chase_database(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    chases = {'main_file_name': json.dumps(_chase.get_main_file_name()),
              'main_file_source': json.dumps(_chase.get_main_file_source()),
              'main_source': json.dumps(_chase.get_main_source()),
              'stack_trace': json.dumps(_chase.get_stack_trace())}

    try:
        chase_group = h5file.create_group('/', 'chases')
        for chase in chases:
            h5file.create_array(chase_group, chase, obj=chases[chase].encode())
        h5file.flush()

    except tables.NodeError:
        raise tables.NodeError('The database has already been chased. ' +
                               'Remove the existing chase prior to ' +
                               '(re)chasing the database.')


def create_database(path, overwrite=True):
    """
    Create a new HDF database that is annotated and chased.

    A new HDF database is created and it is annotated using
    `magni.reproducibility.io.annotate_database` and chased using
    `magni.reproducibility.io.annotate_database`. If the `overwrite` flag is
    true and existing database at `path` is overwritten.

    Parameters
    ----------
    path : str
        The path to the HDF file that is to be created.
    overwrite : bool
        The flag that indicates if an existing database should be overwritten.

    See Also
    --------
    magni.reproducibility.io.annotate_database : Database annotation
    magni.reproducibility.io.chase_database : Database chase

    Examples
    --------
    Create a new database named 'new_db.hdf5':

    >>> from magni.reproducibility.io import create_database
    >>> create_database('new_db.hdf5')

    """

    @_decorate_validation
    def validate_input():
        _generic('path', 'string')
        _numeric('overwrite', 'boolean')

    validate_input()

    if not overwrite and os.path.exists(path):
        raise IOError('{!r} already exists in filesystem.'.format(path))

    with _File(path, mode='w') as h5file:
        annotate_database(h5file)
        chase_database(h5file)


def read_annotations(h5file):
    """
    Read the annotations to an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the annotations are read.

    Returns
    -------
    annotations : dict
        The annotations read from the HDF5 database.

    Raises
    ------
    ValueError
        If the annotations to the HDF5 database does not conform to the Magni
        annotation standard.

    Notes
    -----
    The returned dict holds a key for each annotation in the database. The
    value corresponding to a given key is in itself a dict. See
    `magni.reproducibility.annotate_database` for examples of such annotations.

    Examples
    --------
    Read annotations from the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import read_annotations
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='r') as h5file:
    ...    annotations = read_annotations(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    try:
        h5_annotations = h5file.get_node('/', name='annotations')
    except tables.NoSuchNodeError:
        raise tables.NoSuchNodeError('The database has not been annotated.')

    annotations = dict()
    _recursive_annotation_read(h5_annotations, annotations)

    return annotations


def read_chases(h5file):
    """
    Read the chases to an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the chases are read.

    Returns
    -------
    chasess : dict
        The chases read from the HDF5 database.

    Raises
    ------
    ValueError
        If the chases to the HDF5 database does not conform to the Magni chases
        standard.

    Notes
    -----
    The returned dict holds a key for each chase in the database. The value
    corresponding to a given key is a string. See
    `magni.reproducibility.chase_database` for examples of such chases.

    Examples
    --------
    Read chases from the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import read_chases
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='r') as h5file:
    ...    chases = read_chases(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    try:
        h5_chases = h5file.get_node('/', name='chases')
    except tables.NoSuchNodeError:
        raise tables.NoSuchNodeError('The database has not been chased.')

    h5_chase_dict = h5_chases._v_leaves
    chases = dict()
    try:
        for chase in h5_chase_dict:
            chases[chase] = json.loads(h5_chase_dict[chase].read().decode())
    except ValueError as e:
        raise ValueError('Unable to read the {!r} chase '.format(chase) +
                         'It seems that the chase does not conform to the ' +
                         'Magni chase standard ({!r}).'.format(e.args[0]))

    return chases


def remove_annotations(h5file):
    """
    Remove the annotations from an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the annotations are removed.

    Examples
    --------
    Remove annotations from the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import remove_annotations
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...    remove_annotations(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    try:
        h5file.remove_node('/', 'annotations', recursive=True)
        h5file.flush()
    except tables.NoSuchNodeError:
        pass


def remove_chases(h5file):
    """
    Remove the chases from an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the chases are removed.

    Examples
    --------
    Remove chases from the database named 'db.hdf5':

    >>> import magni
    >>> from magni.reproducibility.io import remove_chases
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...    remove_chases(h5file)

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)

    validate_input()

    try:
        h5file.remove_node('/', 'chases', recursive=True)
        h5file.flush()
    except tables.NoSuchNodeError:
        pass


def write_custom_annotation(h5file, annotation_name, annotation_value,
                            annotations_sub_group=None):
    """
    Write a custom annotation to an HDF5 database.

    The annotation is written to the `h5file` under the `annotation_name` such
    that it holds the `annotation_value`.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database to which the annotation is written.
    annotation_name : str
        The name of the annotation to write.
    annotation_value : a JSON serialisable object
        The annotation value to write.
    annotations_sub_group : str
        The group node under "/annotations" to which the custom annotation is
        written (the default is None which implies that the custom annotation
        is written directly under "/annotations").

    Notes
    -----
    The `annotation_value` must be a JSON seriablisable object.

    Examples
    --------
    Write a custom annotation to an HDF5 database.

    >>> import magni
    >>> from magni.reproducibility.io import write_custom_annotation
    >>> annotation_name = 'custom_annotation'
    >>> annotation_value = 'the value'
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...    write_custom_annotation(h5file, annotation_name, annotation_value)
    ...    annotations = magni.reproducibility.io.read_annotations(h5file)
    >>> str(annotations['custom_annotation'])
    'the value'

    """

    @_decorate_validation
    def validate_input():
        _generic('h5file', tables.file.File)
        _generic('annotation_name', 'string')
        _generic('annotations_sub_group', 'string', ignore_none=True)

    validate_input()

    if annotations_sub_group is not None:
        annotations_group = '/'.join(['/annotations', annotations_sub_group])
    else:
        annotations_group = '/annotations'

    try:
        ann_val = json.dumps(annotation_value)
    except TypeError:
        raise TypeError('The annotation value does not have a valid JSON ' +
                        'representation. It may not be used as an annotation.')

    try:
        h5file.create_array(annotations_group, annotation_name,
                            obj=ann_val.encode(), createparents=True)
        h5file.flush()
    except tables.NodeError:
        raise tables.NodeError(
            'The annotation "{!r}" already exists '.format(annotation_name) +
            'in the database. Remove the old annotation before placing a ' +
            'new one.')


def _recursive_annotation_read(h5_annotations, out_annotations_dict):
    """
    Recursively read annotations from an annotation group

    Parameters
    ----------
    h5_annotations : tables.group.Group
        The group to read annotations from.
    out_annotations_dict : dict
        The dictionary to store the read annotations in.

    """

    leaves = h5_annotations._v_leaves
    subgroups = h5_annotations._v_groups

    # Read leaves
    try:
        for annotation_name, annotation_value in leaves.items():
            out_annotations_dict[annotation_name] = json.loads(
                annotation_value.read().decode())
    except ValueError as e:
        raise ValueError('Unable to read the {!r} '.format(annotation_name) +
                         'annotation. It seems that the annotation ' +
                         'does not conform to the Magni annotation ' +
                         'standard ({!r}).'.format(e.args[0]))

    # Recursively handle subgroups
    for subgroup in subgroups:
        out_annotations_dict[subgroup] = dict()
        _recursive_annotation_read(
            subgroups[subgroup], out_annotations_dict[subgroup])



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing support functionality for the other subpackages.

Routine listings
----------------
multiprocessing
    Subpackage providing intuitive and extensive multiprocessing functionality.
config
    Module providing a robust configger class.
matrices
    Module providing matrix emulators.
plotting
    Module providing utilities for control of plotting using `matplotlib`.
validation
    Subpackage providing validation capability.
types
    Module providing custom data types.
split_path(path)
    Split a path into folder path, file name, and file extension.

Notes
-----
See `_util` for documentation of `split_path`.

"""

# the validation and config modules need to be imported first to avoid
# recursive imports
from magni.utils import validation
from magni.utils import config

from magni.utils import matrices
from magni.utils import multiprocessing
from magni.utils import plotting
from magni.utils import types
from magni.utils._util import split_path



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing an implementation of Iterative Hard Thresholding (IHT).

.. note:: Deprecated in Magni 1.3.0.
          `magni.cs.reconstruction.iht` will be removed in a future version.
          Use the more general `magni.cs.reconstruction.it` instead.

.. warning:: Change of variable interpretation.
             In `magni.cs.reconstruction.it` the config variable
             `threshold_fixed` has a different interpretation than in
             `magni.cs.reconstruction.iht`.

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
run(y, A)
    Run the IHT reconstruction algorithm.

Notes
-----
The IHT reconstruction algorithm is described in [1]_. The default
configuration uses the False Alarm Rate heuristic described in [2]_.

References
----------
.. [1] T. Blumensath and M.E. Davies, "Iterative Thresholding for Sparse
   Approximations", *Journal of Fourier Analysis and Applications*, vol. 14,
   pp. 629-654, Sep. 2008.
.. [2] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
   Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in Signal
   Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.

"""

import warnings

from magni.cs.reconstruction.iht._config import configger as config
from magni.cs.reconstruction import it as _it


def run(y, A):
    """
    Run the IHT reconstruction algorithm.

    .. note:: Deprecated in Magni 1.3.0
          `magni.cs.reconstruction.iht` will be removed in a future version.
          Use the more general `magni.cs.reconstruction.it` instead.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    Examples
    --------
    For example, recovering a vector from random measurements

    >>> import warnings
    >>> import numpy as np
    >>> from magni.cs.reconstruction.iht import run, config
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(80) * np.random.randn(80, 200)
    >>> alpha = np.zeros((200, 1))
    >>> alpha[:10] = 1
    >>> y = A.dot(alpha)
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter('ignore')
    ...     alpha_hat = run(y, A)
    ...
    >>> np.set_printoptions(suppress=True)
    >>> alpha_hat[:12]
    array([[ 0.99836297],
           [ 1.00029086],
           [ 0.99760224],
           [ 0.99927175],
           [ 0.99899124],
           [ 0.99899434],
           [ 0.9987368 ],
           [ 0.99801849],
           [ 1.00059408],
           [ 0.9983772 ],
           [ 0.        ],
           [ 0.        ]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10

    Or recovering the same only using a fixed threshold level:

    >>> config.update({'threshold': 'oracle', 'threshold_fixed': 10./80})
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter('ignore')
    ...     alpha_hat_2 = run(y, A)
    ...
    >>> alpha_hat_2[:12]
    array([[ 0.99877706],
           [ 0.99931441],
           [ 0.9978366 ],
           [ 0.99944973],
           [ 1.00052762],
           [ 1.00033436],
           [ 0.99943286],
           [ 0.99952526],
           [ 0.99941578],
           [ 0.99942908],
           [ 0.        ],
           [ 0.        ]])


    >>> (np.abs(alpha_hat_2) > 1e-2).sum()
    10

    """

    warnings.warn(
        '`magni.cs.reconstruction.iht` is deprecated in  magni 1.3.0. It ' +
        'will be removed in a future version. Use the more general ' +
        '`magni.cs.reconstruction.it` instead.', DeprecationWarning)

    current_it_config = dict(_it.config.items())

    iht_config = dict(config.items())
    k = int(iht_config['threshold_fixed'] * A.shape[0])
    iht_config['threshold_fixed'] = k
    if iht_config['threshold'] == 'oracle':
        iht_config['threshold'] = 'fixed'

    try:
        _it.config.update(iht_config)
        alpha = _it.run(y, A)
    finally:
        _it.config.update(current_it_config)

    return alpha



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for attaching data.

The reading of an .mi file is logically separated into four steps of which the
functionality provided by this module performs the third step.

Routine listings
----------------
attach_data(obj, data)
    Attach data to a hierarchical object-structure.

"""

from __future__ import division

import numpy as np


def attach_data(obj, data):
    """
    Attach data to a hierachical object-structure.

    In the case of .mi image files, the data should be attached to the buffers
    of the file. In the case of .mi spectroscopy files, the data should be
    attached to the chunks of the file.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure to attach data to.
    data : numpy.ndarray
        The 1-dimensional data.

    See Also
    --------
    magni.afm.io.read_mi_file : Function using the present function.

    """

    if obj['attrs']['fileType'].lower() == 'image':
        _attach_image_data(obj, data)
    else:  # 'spectroscopy'
        _attach_spectroscopy_data(obj, data)


def _attach_image_data(obj, data):
    """
    Attach data to a hierarchical object-structure representing an image.

    The data should be attached to the buffers of the file.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure to attach data to.
    data : numpy.ndarray
        The 1-dimensional data.

    See Also
    --------
    attach_data : Function using the present function.

    """

    obj, data, thumbnail = _handle_format_inconsistency(obj, data)
    width, height = obj['attrs']['xPixels'], obj['attrs']['yPixels']
    size = width * height
    index = 0

    for buffer_ in obj['buffers']:
        buffer_['data'] = data[index:index + size]
        buffer_['data'] = buffer_['data'].reshape(height, width)
        buffer_['data'] = buffer_['data'][::-1, :]

        if obj['attrs']['data'].lower() in ('binary', 'binary_32', 'ascii'):
            if 'bufferRange' not in buffer_['attrs']:
                raise IOError("Each buffer must have a 'bufferRange' header "
                              "parameter.")

            if not isinstance(buffer_['attrs']['bufferRange'], float):
                msg = "The 'bufferRange' buffer parameter must have type {!r}."
                raise IOError(msg.format(float))

            buffer_['data'] = buffer_['attrs']['bufferRange'] * buffer_['data']

        index = index + size

    obj['buffers'] = obj['buffers'] + thumbnail

    if data.shape[0] != index:
        raise IOError("The file contains different amount of data than "
                      "specified by the file headers.")


def _attach_spectroscopy_data(obj, data):
    """
    Attach data to a hierarchical object-structure representing a spectroscopy.

    The data should be attached to the chunks of the file.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure to attach data to.
    data : numpy.ndarray
        The 1-dimensional data.

    See Also
    --------
    attach_data : Function using the present function.

    """

    index = 0

    for buffer_ in obj['buffers'][1:]:
        for chunk in buffer_['data']:
            chunk['data'] = data[index:index + chunk['attrs']['samples']]
            index = index + chunk['attrs']['samples']

    if data.shape[0] != index:
        raise IOError("The file contains different amount of data than "
                      "specified by the file headers.")


def _handle_format_inconsistency(obj, data):
    """
    Handle format inconsistency.

    The inconsistency is the optional precense of an undocumented thumbnail.
    If the thumbnail is present, the file header contains a thumbnail parameter
    specifying the resolution of the thumbnail image.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure to attach data to.
    data : numpy.ndarray
        The 1-dimensional data.

    Returns
    -------
    obj : object
        The hierarchical object-structure to attach data to.
    data : numpy.ndarray
        The 1-dimensional data excluding any thumbnail data.
    thumbnail : list
        A list of a thumbnail object-structure if any.

    See Also
    --------
    _attach_image_data : Function using the present function.

    Notes
    -----
    The thumbnail is stored as 8-bit unsigned integers with a red, a green, and
    a blue channel. If the thumbnail is present, it is represented by three
    buffer object-structures; one for each channel.

    """

    if 'thumbnail' in obj['attrs'].keys():
        if not isinstance(obj['attrs']['thumbnail'], str):
            msg = "The 'thumbnail' file header must have type {!r}."
            raise IOError(msg.format(str))

        shape = obj['attrs']['thumbnail'].split('x')

        try:
            shape = (int(shape[0]), int(shape[1]))
        except (IndexError, ValueError):
            raise IOError("The 'thumbnail' file header must consist of two "
                          "integers separated by an 'x'.")

        size = shape[0] * shape[1]

        if obj['attrs']['data'].lower() == 'binary':
            data, thumbnail = data[:-3 * size // 2], data[-3 * size // 2:]
            thumbnail = np.int16(thumbnail * 2**15).view(np.uint8)
        elif obj['attrs']['data'].lower() == 'binary_32':
            data, thumbnail = data[:-3 * size // 4], data[-3 * size // 4:]
            thumbnail = np.int32(thumbnail * 2**31).view(np.uint8)
        else:
            msg = ("The 'thumbnail' file header causes unknown behavior for "
                   "{!r} data.")
            raise IOError(msg.format(obj['attrs']['data']))

        thumbnail = [thumbnail[i * size:(i + 1) * size] for i in range(3)]
        labels = ['ThumbnailRed', 'ThumbnailGreen', 'ThumbnailBlue']
        thumbnail = [{'attrs': {'bufferLabel': labels[i]},
                      'data': np.int16(thumbnail[i].reshape(shape))}
                     for i in range(3)]
    else:
        thumbnail = []

    return obj, data, thumbnail



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of Generalised Approximate Message Passing
(GAMP).

Routine listings
----------------
channel_initialisation
    Module providing functionality for initialisations of GAMP channels.
config
    Configger providing configuration options for this subpackage.
input_channel
    Module providing GAMP input channels.
output_channel
    Module providing GAMP output channels.
run(y, A, A_asq=None)
    Run the GAMP reconstruction algorithm.
stop_criterion
    Module providing GAMP stop criteria.


Notes
-----
Implementations of Mimimum Mean Squared Error (MMSE) Generalised Approximate
Message Passing (GAMP) from [1]_, [2]_ based on description of it in [3]_ are
available. The GAMP is a generalisation of the Approximate Message Passing
(AMP) algorithm derived independelty by Donoho et al. [4]_ and Krzakala et al.
[5]_, [6]_.

This implementation allows custom input- and output channels as well as the use
of sum approximations of the squared system matrix as detailed in [2]_, [5]_.
Furthermore, a simple damping option is available based on the description in
[7]_ (see also [8]_ for more details on damping in GAMP).

References
----------
.. [1] S. Rangan, "Generalized Approximate Message Passing for Estimation with
   Random Linear Mixing", *in IEEE International Symposium on Information
   Theory (ISIT)*, St. Petersburg, Russia, Jul. 31 - Aug. 5, 2011,
   pp. 2168-2172.
.. [2] S. Rangan, "Generalized Approximate Message Passing for Estimation
   with Random Linear Mixing", arXiv:1010.5141v2, pp. 1-22, Aug. 2012.
.. [3] J. T. Parker, "Approximate Message Passing Algorithms for Generalized
   Bilinear Inference", PhD Thesis, Graduate School of The Ohio State
   University, 2014
.. [4] D.L. Donoho, A. Maleki, and A. Montanari, "Message-passing algorithms
   for compressed sensing", *Proceedings of the National Academy of Sciences of
   the United States of America*, vol. 106, no. 45, pp. 18914-18919, Nov. 2009.
.. [5] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
   "Probabilistic reconstruction in compressed sensing: algorithms, phase
   diagrams, and threshold achieving matrices", *Journal of Statistical
   Mechanics: Theory and Experiment*, vol. P08009, pp. 1-57, Aug. 2012.
.. [6] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
   "Statistical-Physics-Based Reconstruction in Compressed Sensing", *Physics
   Review X*, vol. 2, no. 2, pp. (021005-1)-(021005-18), May 2012.
.. [7] S. Rangan, P. Schniter, and A. Fletcher. "On the Convergence of
   Approximate Message Passing with Arbitrary Matrices", *in IEEE International
   Symposium on Information Theory (ISIT)*, pp. 236-240, Honolulu, Hawaii, USA,
   Jun. 29 - Jul. 4, 2014.
.. [8] J. Vila, P. Schniter, S. Rangan, F. Krzakala, L. Zdeborova, "Adaptive
   Damping and Mean Removal for the Generalized Approximate Message Passing
   Algorithm", *in IEEE International Conference on Acoustics, Speech, and
   Signal Processing (ICASSP)*, South Brisbane, Queensland, Australia, Apr.
   19-24, 2015, pp. 2021-2025.

"""

from magni.cs.reconstruction.gamp import (
    channel_initialisation, input_channel, output_channel, stop_criterion)
from magni.cs.reconstruction.gamp._config import configger as config
from magni.cs.reconstruction.gamp._algorithm import run



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.iht`
subpackage.

See also
--------
magni.cs.reconstruction._config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

iterations : int
    The maximum number of iterations to do (the default is 300).
kappa_fixed : float
    The relaxation parameter used in the algorithm (the default is 0.65).
precision_float : {np.float, np.float16, np.float32, np.float64, np.float128}
    The floating point precision used for the computations (the default is
    np.float64).
threshold : {'far', 'oracle'}
    The method for selecting the threshold value.
threshold_fixed : float
    The assumed rho value used for selecting the threshold value if using the
    oracle method.
tolerance : float
    The least acceptable ratio of residual to measurements (in 2-norm) to break
    the interations (the default is 0.001).

"""

from __future__ import division

import numpy as np

from magni.cs.reconstruction._config import Configger as _Configger
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


configger = _Configger(
    {'iterations': 300,
     'kappa_fixed': 0.65,
     'precision_float': np.float64,
     'threshold': 'far',
     'threshold_fixed': 0.1,
     'tolerance': 1e-3},
    {'iterations': _numeric(None, 'integer', range_='[1;inf)'),
     'kappa_fixed': _numeric(None, 'floating', range_='[0;1]'),
     'precision_float': _generic(None, type, value_in=(
         np.float,
         getattr(np, 'float16', np.float_),
         getattr(np, 'float32', np.float_),
         getattr(np, 'float64', np.float_),
         getattr(np, 'float128', np.float_))),
     'threshold': _generic(None, 'string', value_in=('far', 'oracle')),
     'threshold_fixed': _numeric(None, 'floating', range_='[0;1]'),
     'tolerance': _numeric(None, 'floating', range_='[0;inf]')})



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the core Smoothed l0 (SL0) algorithm.

Routine listings
----------------
run(y, A)
    Run the SL0 reconstruction algoritm.

See Also
--------
magni.cs.reconstruction.sl0._config : Configuration options.

Notes
-----
Implementations of the original SL0 reconstruction algorithm [1]_ and a
modified Sl0 reconstruction algorithm [3]_ are available. It is also possible
to configure the subpackage to provide customised versions of the SL0
reconstruction algorithm. The projection algorithm [1]_ is used for small delta
(< 0.55) whereas the contraint elimination algorithm [2]_ is used for large
delta (>= 0.55) which merely affects the computation time.

References
----------
.. [1] H. Mohimani, M. Babaie-Zadeh, and C. Jutten, "A Fast Approach for
   Overcomplete Sparse Decomposition Based on Smoothed l0 Norm", *IEEE
   Transactions on Signal Processing*, vol. 57, no. 1, pp. 289-301, Jan. 2009.
.. [2] Z. Cui, H. Zhang, and W. Lu, "An Improved Smoothed l0-norm Algorithm
   Based on Multiparameter Approximation Function", *in 12th IEEE International
   Conference on Communication Technology (ICCT)*, Nanjing, China, Nov. 11-14,
   2011, pp. 942-945.
.. [3] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from __future__ import division

import numpy as np
import scipy.linalg

from magni.cs.reconstruction.sl0 import config as _conf
from magni.cs.reconstruction.sl0 import _L_start
from magni.cs.reconstruction.sl0 import _L_update
from magni.cs.reconstruction.sl0 import _mu_start
from magni.cs.reconstruction.sl0 import _mu_update
from magni.cs.reconstruction.sl0 import _sigma_start
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


def run(y, A):
    """
    Run the SL0 reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    See Also
    --------
    magni.cs.reconstruction.sl0.config : Configuration options.

    Notes
    -----
    The algorithm terminates after a fixed number of iterations or if the ratio
    between the 2-norm of the residual and the 2-norm of the measurements falls
    below the specified `tolerance`.

    Examples
    --------
    For example, recovering a vector from random measurements using the
    original SL0 reconstruction algorithm

    >>> import numpy as np, magni
    >>> from magni.cs.reconstruction.sl0 import run
    >>> np.set_printoptions(suppress=True)
    >>> magni.cs.reconstruction.sl0.config['L'] = 'fixed'
    >>> magni.cs.reconstruction.sl0.config['mu'] = 'fixed'
    >>> magni.cs.reconstruction.sl0.config['sigma_start'] = 'fixed'
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(80) * np.random.randn(80, 200)
    >>> alpha = np.zeros((200, 1))
    >>> alpha[:10] = 1
    >>> y = A.dot(alpha)
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:12]
    array([[ 0.99993202],
           [ 0.99992793],
           [ 0.99998107],
           [ 0.99998105],
           [ 1.00005882],
           [ 1.00000843],
           [ 0.99999138],
           [ 1.00009479],
           [ 0.99995889],
           [ 0.99992509],
           [-0.00001509],
           [ 0.00000275]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10

    Or recover the same vector as above using the modified SL0 reconstruction
    algorithm

    >>> magni.cs.reconstruction.sl0.config['L'] = 'geometric'
    >>> magni.cs.reconstruction.sl0.config['mu'] = 'step'
    >>> magni.cs.reconstruction.sl0.config['sigma_start'] = 'reciprocal'
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:12]
    array([[ 0.9999963 ],
           [ 1.00000119],
           [ 1.00000293],
           [ 0.99999661],
           [ 1.00000021],
           [ 0.9999951 ],
           [ 1.00000103],
           [ 1.00000662],
           [ 1.00000404],
           [ 0.99998937],
           [-0.00000075],
           [ 0.00000037]])
    >>> (np.abs(alpha_hat) > 1e-2).sum()
    10

    """

    @_decorate_validation
    def validate_input():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('A', ('integer', 'floating', 'complex'),
                 shape=(y.shape[0], -1))

    @_decorate_validation
    def validate_output():
        _numeric('alpha', ('integer', 'floating', 'complex'),
                 shape=(A.shape[1], 1))

    validate_input()

    if not isinstance(A, np.ndarray):
        A = A.A

    param = dict(_conf.items())
    convert = param['precision_float']
    epsilon = param['epsilon']
    sigma_geometric = param['sigma_geometric']
    sigma_stop = param['sigma_stop_fixed']

    L = _L_start.get_function_handle(param['L'], locals())()
    mu = _mu_start.get_function_handle(param['mu'], locals())()
    sigma_start = _sigma_start.get_function_handle(
        param['sigma_start'], locals())()

    calculate_L_update = _L_update.get_function_handle(param['L'], locals())
    calculate_mu_update = _mu_update.get_function_handle(param['mu'], locals())

    if A.shape[0] / A.shape[1] < 0.55:
        Q, R = scipy.linalg.qr(A.T, mode='economic')
        A_pinv = Q.dot(scipy.linalg.inv(R.T))
        alpha = A_pinv.dot(y)
    else:
        Q, R = scipy.linalg.qr(A.T)
        Q1 = Q[:, :A.shape[0]]
        Q2 = Q[:, A.shape[0]:]
        R = R[:R.shape[1], :]
        alpha = Q1.dot(scipy.linalg.solve_triangular(R, y, trans='T'))

    sigma = convert(sigma_start) * np.abs(alpha).max()

    alpha_zeros = np.zeros(alpha.shape)
    i = 0

    while sigma > sigma_stop:
        alpha_prev = alpha_zeros
        j = 0

        while (scipy.linalg.norm(alpha - alpha_prev) > sigma * epsilon and
               j <= L):
            alpha_prev = alpha.copy()
            d = np.exp(-alpha**2 / (2 * sigma**2)) * alpha

            if A.shape[0] / A.shape[1] < 0.55:
                alpha = alpha - mu * d
                alpha = alpha - A_pinv.dot(A.dot(alpha) - y)  # Projection
            else:
                nabla = Q2.T.dot(d)
                alpha = alpha - Q2.dot(mu * nabla)  # Search on feasible set

            j = j + 1

        sigma = sigma * sigma_geometric
        L = calculate_L_update(locals())
        mu = calculate_mu_update(locals())
        i = i + 1

    validate_output()

    return alpha
